# coding: utf8
""" This module defines EvtDataFrame, a class which methods are
    designed to process input data from several formats. It also
    define Store, a class designed to store time series values
    in HDF5 format. """

from __future__ import print_function
import sys
import numpy as np
import pandas as pd
import evaltools as evt
from datetime import datetime, date, timedelta, time
from os.path import isdir, isfile
import h5py


class EvtDataFrame(object):
    """ DataFrame class for evaltools input data.

    This class is based on pandas.DataFrame class. The main attribute
    (DF) is a pandas DataFrame with datetime index and a list of
    stations as columns.

    Parameters
    ----------
    stations : 1D array of str
        List of stations to keep in the returned EvtDataFrame.
    startDate : datetime.date object
        The date from which data is collected.
    endDate : datetime.date object
        The data until which data is collected.
    name : str
        Name for the dataframe.
    species : str
        Species to process.

    """
    def __init__(self, stations, startDate, endDate, species="", seriesType='hourly'):
        # check arguments integrity
        if startDate > endDate:
            raise evt.EvaltoolsError("startDate = {startDate} must be <= endDate = {endDate}".format(startDate=startDate, endDate=endDate))

        # initialize attributes
        self.species = species
        self.nb_days = (endDate-startDate).days + 1
        self.startDate = startDate
        self.endDate = endDate

        # initialize dataframe according to the series type
        self.seriesType = seriesType
        if seriesType == 'hourly':
            startTime = datetime.combine(startDate,time(0))
            times = pd.DatetimeIndex([startTime+timedelta(hours=h) for h in range(self.nb_days*24)])
        elif seriesType == 'daily':
            times = pd.DatetimeIndex([self.startDate+timedelta(days=d) for d in range(nb_days)])
        else:
            raise evt.EvaltoolsError("seriesType argument must be either 'hourly' or 'daily' !!!")
        self.DF = pd.DataFrame(index=times, columns=stations, dtype=float)

    def nan_rate(self):
        """ Computed rate of missing values on in the whole dataframe. """
        nan_rate = np.sum(np.sum(self.DF.isna())) / ((self.DF.shape[0]*self.DF.shape[1])*1.)
        return nan_rate

    def summary(self):
        """ Print summary statistics on the data.

        Nan rate, minimum and maximum are computed for each station. Then,
        the minimum, maximum and median of each of these statistics are
        displayed.

        """
        nanRate = (self.DF.isna().sum()/self.DF.shape[0]).quantile([0, .5, 1])
        min_val = self.DF.min().quantile([0, .5, 1])
        max_val = self.DF.max().quantile([0, .5, 1])

        summary = pd.DataFrame({'nan rate': nanRate, 'min value': min_val, 'max value': max_val})
        summary.index.name = 'quantiles'
        print(summary)

    def subPeriod(self, startDate, endDate):
        """ Build a new EvtDataFrame object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            EvtDataFrame object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError("startDate must be >= {} !!!".format(self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError("endDate must be <= {} !!!".format(self.endDate))

        evtDF = self.__class__(stations=self.DF.columns, startDate=startDate,
            endDate=endDate, species=self.species, seriesType=self.seriesType)
        evtDF.DF.update(self.DF)

        return evtDF

    def drop_unrepresentative_stations(self, availability_ratio=0.75):
        """ Drop stations with a certain rate of missing values.

        Modify EvtDataFrame object in place.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available on the period required to
            keep a station.

        """
        if availability_ratio is None:
            idx = (np.sum(~self.DF.isna()) == 0.)  # index of stations to be dropped
        else:
            idx = np.sum(~self.DF.isna())/self.DF.shape[0] < availability_ratio  # index of stations to be dropped
        dropped_sta = idx.index[idx]
        self.DF.drop(columns=dropped_sta, inplace=True)

    def read_EEA_obs_csv(self, generic_file_path, correc_unit=1):
        """ Collect data from EEA observations csv files.

        Modify EvtDataFrame.DF in place using non-NA values from EEA
        observations csv files.

        Parameters
        ----------
        generic_file_path : str
            Path of the files with {year}, {month}, {day} instead
            of the year, month and day number, and {species} instead
            instead of the species name.
        correc_unit : float
            Multiplicative factor applied to original values.

        """
        if self.seriesType != 'hourly':
            raise evt.EvaltoolsError("EvtDataFrame series type must be 'hourly' to use read_EEA_obs_csv method !!!")

        for d in range(-1, self.nb_days):  # -1 since an EEA csv contains 1, 2, ..., 23 , 0 (of D+1)
            currentDate = self.startDate+timedelta(days=d)
            f_path = generic_file_path.format(species=self.species, year=currentDate.year, month=currentDate.strftime('%m'), day=currentDate.strftime('%d'))
            # print(f_path)
            if isfile(f_path):
                df = pd.read_table(f_path, sep=';', names=['station', 'time', 'obs'],  # names=['time', 'lat', 'lon', 'class', 'num_param', 'obs', 'altitude', 'sflag'],
                                   # index_col=0,
                                   header=None,  # usecols=['time', 'obs'])
                                   usecols=[0, 1, 6])
                df.time = pd.to_datetime(df.time, format='%Y%m%d%H%M%S')
                df.drop_duplicates(inplace=True)
                try:
                    df = df.pivot(index='time', columns='station', values='obs')
                    if correc_unit != 1:
                        df = df*correc_unit
                    self.DF.update(df)
                except:
                    print("*** Duplicated station-datetime couple with " \
                          "different values in file {}: this file is skipped".format(f_path))
            else:
                print("*** Missing file: {}".format(f_path))

    def updateFromTimeSeries(self, generic_file_path, correc_unit=1):
        """ Update nan values of the current object with timeseries files.

        Parameters
        ----------
        generic_file_path : str
            Generic path of timeseriesfiles with {station} instead
            of the station name and {forecastDay}
        correc_unit : float
            Multiplicative factor applied to original values.

        """
        for sta in self.DF.columns:
            file_path = generic_file_path.format(station=sta)
            if isfile(file_path):
                df = evt.timeseries.readtimeseries(start_date=self.startDate,
                    end_date=self.endDate,
                    lag=0,
                    lfiles=[file_path],
                    correc_unit=correc_unit,
                    seriesType=self.seriesType)
                df.columns = [sta]
                self.DF.update(df)

    def updateFromEvtDF(self, evtDF, correc_unit=1):
        """ Update nan values of the current object with values from an other EvtDataFrame.

        Parameters
        ----------
        evtDF : evaltools.dataframe.EvtDataFrame
            EvtDataFrame where to take data.
        correc_unit : float
            Multiplicative factor applied to original values.

        """
        try:
            self.DF.update(evtDF.DF*correc_unit)
        except AttributeError:
            self.DF.update(evtDF*correc_unit)


def deltaToolFormatting(EvtDataFrame_dict, outputDir):
    """ """
    station_dataframes = {}
    for lab, edf in EvtDataFrame_dict.iteritems():
        for sta in edf.DF.columns:
            if sta in station_dataframes.keys():
                station_dataframes[sta][lab] = pd.DataFrame(edf.DF[sta].astype(float))
            else:
                station_dataframes[sta] = pd.DataFrame(index=edf.DF.index, columns=['year', 'month', 'day', 'hour', lab])
                station_dataframes[sta].year = edf.DF.index.year
                station_dataframes[sta].month = edf.DF.index.month
                station_dataframes[sta].day = edf.DF.index.day
                station_dataframes[sta].hour = edf.DF.index.hour
                station_dataframes[sta][lab] = pd.DataFrame(edf.DF[sta].astype(float))
    for sta, df in station_dataframes.iteritems():
        station_dataframes[sta].to_csv("{outputDir}/{station}.csv".format(outputDir=outputDir, station=sta), sep=';', na_rep='-999', float_format='%.5g', header=True, index=False)


class Store(object):
    """ Tool designed for storing hourly time series data in HDF5 files.

    A file created with this class has a special architecture. It can
    hold data from different species on a same time period. The time
    period covered has a fixed length but can be shifted day by day.
    Boundary dates are saved in the file attributes.

    For each species, several dataset can be stored containing data
    for a same station list that can be increased.

    Parameters
    ----------
    file_path : str
        Path of the hdf5 file to read. The file must exist.

    """
    char_dtype = h5py.special_dtype(vlen=str)

    def __init__(self, file_path, read_only=False):
        # open file
        if read_only is True:
            self.file = h5py.File(file_path, 'r')
        else:
            self.file = h5py.File(file_path, 'r+')

        # load dates
        self.startDate = date(int(self.file.attrs['start'][0:4]), int(self.file.attrs['start'][4:6]), int(self.file.attrs['start'][6:8]))
        self.endDate = date(int(self.file.attrs['end'][0:4]), int(self.file.attrs['end'][4:6]), int(self.file.attrs['end'][6:8]))
        self.nb_days = int((self.endDate - self.startDate).days + 1)

        try:
            self.seriesType = self.file.attrs['seriesType'].decode()
        except:
            self.file.attrs['seriesType'] = 'hourly'
            self.seriesType = self.file.attrs['seriesType']

        # nTimes is a number of hours (seriesType == 'hourly') or days (seriesType == 'daily')
        if self.seriesType == 'hourly':
            self.nTimes = self.nb_days*24
        elif self.seriesType == 'daily':
            self.nTimes = self.nb_days
        else:
            raise evt.EvaltoolsError("seriesType argument must be either 'hourly' or 'daily' !!!")

    @classmethod
    def newFile(cls, file_path, startDate, endDate, seriesType='hourly'):
        """ Build a Store object by writing a new hdf5 file.

        If the file already exists, it is truncated.

        Parameters
        ----------
        file_path : str
            Path of the file to create.
        startDate : datetime.date
            Starting date of the file data.
        endDate : datetime.date
            Ending date of the file data.

        Returns
        -------
            New Store object.

        """
        new_file = h5py.File(file_path, 'w')
        new_file.attrs['start'] = startDate.strftime('%Y%m%d')
        new_file.attrs['end'] = endDate.strftime('%Y%m%d')
        new_file.attrs['seriesType'] = seriesType
        return cls(file_path)

    def summary(self):
        """ Print summary statistics on the hdf5 file data.

        Nan rate, minimum and maximum are computed for each station. Then,
        the minimum, maximum and median of each of these statistics are
        displayed.

        """
        print("File = {}".format(self.file))
        print("Period = {startDate} - {endDate}".format(startDate=self.startDate, endDate=self.endDate))
        print("Time step = {}".format(self.seriesType))
        for species in self.file.keys():
            print("")
            print("---------- {} ----------".format(species))
            print("number of stations = {}".format(self.file[species]['stations']['code'].shape[0]))
            for model in self.file[species]['data'].keys():
                print("")
                print("* {} *".format(model))
                evtDF = self.getDF(species=species, dsetName=model, startDate=self.startDate, endDate=self.endDate)
                evtDF.summary()

    def newSpecies(self, species, stations, dsetNames=[]):
        """ Add a new species to the file of the Store object.

        Parameters
        ----------
        species : str
            Species name that will be the name of the new hdf5 file group.
        stations : list of str
            List of the station codes corresponding to the columns of the
            datasets to be stored.
        dsetNames : None or list of str
            List of the dataset names to initialize. The new dataset will be
            filled with np.nan.

        """
        grp = self.file.create_group(species)
        grp.create_group('stations')
        grp['stations'].create_dataset('code', data=np.array(stations, dtype=self.char_dtype), maxshape=(None,))
        grp.create_group('data')
        for name in dsetNames:
            self.newDataset(species, name)

    def newDataset(self, species, dsetName):
        """ Add a new dataset to the file of the Store object.

        Parameters
        ----------
        species : str
            Species name that refered to the file group name
            where to add the dataset.
        dsetName : str
            Name of the new dataset. The new dataset will be
            filled with np.nan.

        """
        if species not in self.file.keys():
            raise evt.EvaltoolsError("{} is not a species of the hdf5 file. Please use newSpecies method to add it.".format(species))

        grp = self.file[species]
        grp['data'].create_dataset(dsetName, dtype='float32', 
            data=np.full((self.nTimes, grp['stations']['code'].shape[0]), np.nan),
            compression="gzip", compression_opts=9, maxshape=(self.nTimes, None),
            fillvalue=np.nan)

    def addStations(self, species, new_stations):
        """ Add stations to the species dataframes.

        Add columns filled with np.nan to all the dataset
        of the species.
        New station codes must not already be present.

        Parameters
        ----------
        species : str
            Species name that refered to the file group name
            where to add the new stations.
        new_stations : list of str
            List of the codes of the new stations.

        """
        if np.isin(new_stations, self.file[species]['stations']['code'][:]).any():
            raise evt.EvaltoolsError("Trying to add already existing stations !!!")

        stations = np.concatenate((self.file[species]['stations']['code'][:], new_stations))
        nsta = stations.shape[0]
        self.file[species]['stations']['code'].resize((nsta,))
        self.file[species]['stations']['code'][len(stations)-len(new_stations):] = np.array(new_stations).astype('|S7')

        for dset in self.file[species]['data'].values():
            dset.resize((self.nTimes, nsta))

    def update(self, evtDataFrame, dsetName, add_new_stations=True):
        """ Update a dataset of the hdf5 file with values from an EvtDataFrame object.

        Modify a dataset of the hdf5 file using non-NA values from passed
        EvtDataFrame object. The species of the modified dataset is the
        species of the EvtDataFrame object.

        Parameters
        ----------
        evtDataFrame : EvtDataFrame object
            EvtDataFrame object containing new values to add to the file.
        dsetName : str
            Name of the dataset to update.
        add_new_stations : bool
            If true, stations only present in evtDataFrame are also
            added to the hdf5 file.

        """
        # check if series types match
        if self.seriesType != evtDataFrame.seriesType:
            raise evt.EvaltoolsError("Series types of evtDataFrame and self must be equal !!!")

        # check if h5 file holds the appropriate dataset
        species = evtDataFrame.species
        if species not in self.file.keys():
            raise evt.EvaltoolsError("{} is not a species of the hdf5 file. Please " \
                "use newSpecies method to add it.".format(species))

        if dsetName not in self.file[species]['data'].keys():
            raise evt.EvaltoolsError("{dset} is not a dataset name of {species} " \
                "group of the hdf5 file. Please use newDataset method to " \
                "add it.".format(dset=dsetName, species=species))

        # find indices of data to load from the hdf5 file
        if self.startDate > evtDataFrame.startDate:
            first_idx = 0
            startTime = datetime.combine(self.startDate, time(0))
        else:
            first_idx = (evtDataFrame.startDate - self.startDate).days*24
            startTime = datetime.combine(evtDataFrame.startDate, time(0))

        if self.endDate < evtDataFrame.endDate:
            last_idx = self.nTimes
        else:
            last_idx = self.nTimes - (self.endDate - evtDataFrame.endDate).days*24

        # get station list from h5 file and add some if missing
        stations = self.file[species]['stations']['code'][:]
        if add_new_stations:
            new_stations = evtDataFrame.DF.columns[np.logical_not(np.isin(evtDataFrame.DF.columns, stations))]
            self.addStations(species, new_stations)
            stations = self.file[species]['stations']['code'][:]

        # construct datetime index
        if self.seriesType == 'hourly':
            times = pd.DatetimeIndex([startTime+timedelta(hours=h) for h in range(last_idx-first_idx)])
        elif self.seriesType == 'daily':
            times = pd.DatetimeIndex([startTime.date()+timedelta(days=d) for d in range(last_idx-first_idx)])

        # get h5 file dataset
        df = pd.DataFrame(self.file[species]['data'][dsetName][first_idx:last_idx, :],
                          index=times,
                          columns=stations)

        # write updated dataset in the h5 file
        df.update(evtDataFrame.DF)
        self.file[species]['data'][dsetName][first_idx:last_idx, :] = df.values

    def shiftOneDay(self):
        """ Shift the hdf5 file time period one day.

        In all datasets of the file, rows corresponding to the
        earliest day are removed and new rows containing np.nan
        are addes for the new day.

        """
        for spe in self.file.keys():
            grp = self.file[spe]['data']
            for model in grp.keys():
                arr = grp[model][24:(24*self.nb_days)]
                grp[model][0:(24*(self.nb_days-1))] = arr
                grp[model][(24*(self.nb_days-1)):(24*self.nb_days)] = np.nan
        self.startDate = self.startDate + timedelta(days=1)
        self.endDate = self.endDate + timedelta(days=1)
        self.file.attrs['start'] = self.startDate.strftime('%Y%m%d')
        self.file.attrs['end'] = self.endDate.strftime('%Y%m%d')

    def getDF(self, species, dsetName, startDate, endDate):
        """ Get a dataset of the hdf5 file.

        Parameters
        ----------
        species : str
            Species name (corresponding to the first wanted key of
            the hdf5 file).
        dsetName : str
            Name of the dataset to get.
        startDate : datetime.date object
            The date from which data is collected.
        endDate : datetime.date object
            The data until which data is collected.

        Returns
        -------
            evaltools.dataframe.EvtDataFrame

        """
        # find indices of data to load from the hdf5 file
        if self.startDate > startDate:
            first_idx = 0
            startTime = datetime.combine(self.startDate, time(0))
        else:
            first_idx = (startDate - self.startDate).days*(24*(self.seriesType == 'hourly')+1*(self.seriesType == 'daily'))
            startTime = datetime.combine(startDate, time(0))
        if self.endDate < endDate:
            last_idx = self.nTimes
        else:
            last_idx = self.nTimes - (self.endDate - endDate).days*(24*(self.seriesType == 'hourly')+1*(self.seriesType == 'daily'))

        # construct datetime index
        if self.seriesType == 'hourly':
            times = pd.DatetimeIndex([startTime+timedelta(hours=h) for h in range(last_idx-first_idx)])
        elif self.seriesType == 'daily':
            times = pd.DatetimeIndex([startTime.date()+timedelta(days=d) for d in range(last_idx-first_idx)])

        # get DataFrame
        df = pd.DataFrame(self.file[species]['data'][dsetName][first_idx:last_idx, :],
                          index=times,
                          columns=self.file[species]['stations']['code'])

        # construct EvtDataFrame
        res = EvtDataFrame(stations=df.columns, startDate=startDate, endDate=endDate, species=species, seriesType=self.seriesType)
        res.DF.update(df)

        return res
