# coding: utf8

from __future__ import print_function

import numpy as np
import pandas as pd
import evaltools as evt
from datetime import datetime, date, timedelta, time
from os.path import isdir, isfile
import pickle
import copy
from functools import reduce


class Observations(object):
    """ Class gathering observations of the studied case.

    An object of class Observations will be specific to one species,
    one series type and one period.

    Parameters
    ----------
    species : str
        Species name (ex: "o3").
    start : datetime.date
        Starting day of the studied period.
    end : datetime.date
        Ending day (included) of the studied period.
    stations : pandas.DataFrame
        DataFrame with station names as index.
    seriesType : str. It can be 'hourly' or 'daily'.
    obsDF : pandas.DataFrame
        DataFrame with datetime index from 'start' to 'end'+'forecastHorizon'-1
        and containing one column per station.
    forecastHorizon : int
        Number of day corresponding to the forecast horizon of the model.
        For example, if the forcast horizon is 4 days, the end date of the studied observations
        has to be 3 days further than the end of the studied period since the studied period
        corresponds to the period along which the model has been executed.

    """
    def __init__(self, species, start, end, stations, seriesType, obsDF, forecastHorizon=1):
        self.species = species
        if type(start) is date:
            self.startDate = start
            self.endDate = end
        else:
            self.startDate = date(int(start[0:4]), int(start[4:6]), int(start[6:8]))
            self.endDate = date(int(end[0:4]), int(end[4:6]), int(end[6:8]))
        self.stations = stations
        self.forecastHorizon = forecastHorizon
        self.seriesType = seriesType

        # initialize obsDF
        self.obsDF = obsDF
        self.obsDF.index = pd.to_datetime(self.obsDF.index)

        # check is stations well corresponds to obsDF columns
        if not np.isin(obsDF.columns, stations.index).all():
            raise evt.EvaltoolsError("obsDF.columns must be contained in stations.index !!!")

        # check if start/end well correspond to obsDF index
        nb_days = (self.endDate-self.startDate).days + forecastHorizon
        if seriesType == 'hourly':
            start_time = datetime.combine(self.startDate,time(0))
            steps = [start_time+timedelta(hours=h) for h in range(nb_days*24)]
        elif seriesType == 'daily':
            start_time = self.startDate
            steps = [start_time+timedelta(days=d) for d in range(nb_days)]
        else:
            raise evt.EvaltoolsError("seriesType argument must be either 'hourly' or 'daily' !!!")
        obsDF = obsDF.loc[obsDF.index.isin(steps)]
        try:
            assert((steps == obsDF.index).all())
        except:
            raise evt.EvaltoolsError("obsDF index do not fit start/end arguments !!!")


    @classmethod
    def fromTimeSeries(cls, generic_file_path, species, start, end, stations,
        correc_unit=1, seriesType='hourly', forecastHorizon=1):
        """ Class method used to construct an object from timeseries files.

        Parameters
        ----------
        generic_file_path : str
            Generic path of timeseriesfiles with {year} instead of
            the year number and {station} instead of the station name.
        species : str
            Species name (ex: "o3").
        start : datetime.date
            Starting day of the studied period.
        end : datetime.date
            Ending day (included) of the studied period.
        stations : pandas.DataFrame
            DataFrame with station names as index.
        correc_unit : float
            multiplicative factor applied to original values.
        seriesType : str
            Must be equal to 'hourly' if time series files contain one
            row per hour, or equal to 'daily' if they contain one row
            per day.
        forecastHorizon : int
            Number of day corresponding to the forecast horizon of the model.
            For example, if the forcast horizon is 4 days, the end date of
            the studied observations has to be 3 days further than the end
            of the studied period since the studied period corresponds to
            the period along which the model has been executed.

        Returns
        -------
            evaltools.evaluator.Observations object

        """
        obsDF = evt.timeseries.get_DF(stationsIdx=stations.index,
                            generic_file_path=generic_file_path,
                            start_date=start,
                            end_date=end+timedelta(days=forecastHorizon-1),
                            lag=0,
                            correc_unit=correc_unit,
                            seriesType=seriesType)
        obj = cls(seriesType=seriesType, obsDF=obsDF, species=species, start=start,
                  end=end, stations=stations, forecastHorizon=forecastHorizon) 
        obj.stations = obj.stations.loc[obj.obsDF.columns]
        return obj

    @classmethod
    def fromEvtDataFrame(cls, edf, listing_path, forecastHorizon=1, correc_unit=1):
        """ Class method used to construct an object from an evaltools.dataframe.EvtDataFrame object.

        Parameters
        ----------
        edf : evaltools.dataframe.EvtDataFrame object
            EvtDataFrame object corresponding observation data.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.

        Returns
        -------
            evaltools.evaluator.Observations object

        """
        stations = evt.utils.read_listing(listing_path)
        if not np.isin(edf.DF.columns, stations.index).all():
            raise evt.EvaltoolsError("Listing does not contain all stations of the EvtDataFrame !!!")
        stations = stations.loc[edf.DF.columns]
        obsDF = edf.DF
        obj = cls(species=edf.species, start=edf.startDate,
                  end=edf.endDate-timedelta(days=forecastHorizon-1),
                  stations=stations, seriesType=edf.seriesType,
                  obsDF=obsDF,
                  forecastHorizon=forecastHorizon)
        if correc_unit != 1:
            obj.obsDF = obj.obsDF*correc_unit
        return obj

    @classmethod
    def fromHDF(cls, startDate, endDate, species, hdf_path, listing_path, forecastHorizon=1, correc_unit=1):
        """ Class method used to construct an object from HDF5 files.

        Parameters
        ----------
        startDate : datetime.date
            Starting day of the studied period.
        endDate : datetime.date
            Ending day (included) of the studied period.
        species : str
            Species name (must be a key of the hdf file).
        hdf_path : str
            Path of the input file. Input data can be splitted over
            two hdf files corresponding to two different years. In
            that case, the file path must contain '{year}'.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.

        Returns
        -------
            evaltools.evaluator.Observations object

        """
        store_list = [evt.dataframe.Store(hdf_path.format(year=year), read_only=True) for year in np.unique([startDate.year, endDate.year])]
        edf_list = [store.getDF(species=species, dsetName='observations', startDate=startDate, endDate=endDate+timedelta(days=forecastHorizon-1)) for store in store_list]
        obs_evtDF = evt.dataframe.EvtDataFrame(stations=reduce(np.union1d, [d.DF.columns for d in edf_list]),
            startDate=startDate, endDate=endDate+timedelta(days=forecastHorizon-1), species=species)
        for edf in edf_list:
            obs_evtDF.updateFromEvtDF(edf)
        observations = evt.evaluator.Observations.fromEvtDataFrame(obs_evtDF, listing_path=listing_path, forecastHorizon=forecastHorizon, correc_unit=correc_unit)

        return observations

    @classmethod
    def from_csv(cls, inputFile, seriesType, species, start, end, stations, forecastHorizon=4):
        """ Class method used to construct an object from a csv file.

        Parameters
        ----------
        inputFile : str
            Path of csv file corresponding to Observations._obsDF attribute.
        seriesType : str
            It can be 'hourly' or 'daily'.
        species : str. Species name (ex: "o3").
        start : str or datetime.date
            Start day of the studied period. If str, it must
            be of the form "yyyymmdd".
        end : str or datetime.date
            End day (included) of the studied period. If str, it must
            be of the form "yyyymmdd".
        stations : pandas.DataFrame
            DataFrame with station names as index.
        forecastHorizon : int
            Number of day corresponding to the forcast horizon of the model.
            For example, if the forcast horizon is 4 days, the end date of the studied observations
            has to be 3 days further than the end of the studied period since the studied period
            corresponds to the period along which the model has been executed.

        """
        if not isfile(inputFile):
            raise evt.EvaltoolsError("Missing file : {0} !!!".format(inputFile))
        obj = cls(species, start, end, stations, seriesType, forecastHorizon=forecastHorizon)
        obj.obsDF = pd.read_csv(inputFile, sep=' ', index_col=0)
        obj.obsDF.index = pd.to_datetime(obj.obsDF.index, format='%Y%m%d%H')
        obj.stations = obj.stations.loc[obj.obsDF.columns]
        return obj

    def to_csv(self, filePath):
        """ Save timeseries dataframe as csv.

        Parameters
        ----------
        filePath : str
            Csv file path.

        """
        filePath = filePath
        self.obsDF.to_csv(filePath, sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')

    def check_values(self, threshold, drop=False, filePath=None):
        """ Check if observation values exceed a threshold.

        If there are values above the threshold, a message is printed
        and these values are set to nan if drop == True.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        drop : bool
            If True, values above the threshold are set to nan.
        filePath : None or str
            File path where to save the names of stations that exceed
            the threshold.

        """
        idx = (self.obsDF > threshold)
        nb_val_above = np.sum(idx.values)
        if nb_val_above > 0:
            print("*** {nb} values above threshold in observations !!! values are {action}.".format(nb=nb_val_above, action='kept'*(not drop)+'set to np.nan'*drop))
            if filePath is not None:
                self.obsDF.stack().loc[idx.stack()].to_csv(filePath, sep=' ', na_rep='nan', float_format='%g', header=False, index=True, date_format='%Y%m%d%H')
            if drop is True:
                self.obsDF[idx] = np.nan
        else:
            print("0 value above threshold in observations")

    def persistenceModel(self, color='k'):
        """ Build an Evaluator object simulation data based on persistence model.

        Parameters
        ----------
        color : str
            Default color that will be used in plotting functions
            for the new object.

        Returns
        -------
            Evaluator object.

        """
        simDF = []
        for fday in range(self.forecastHorizon):
            df = self.obsDF.loc[np.logical_and(self.obsDF.index.date >= self.startDate,
                                               self.obsDF.index.date <= self.endDate-timedelta(days=1))]
            simDF.append(pd.DataFrame(np.array(df), columns=df.columns, index=df.index+timedelta(days=fday+1)))
        simulations = Simulations(species=self.species, model="persistence",
            start=self.startDate+timedelta(days=1), end=self.endDate, seriesType=self.seriesType,
            simDF=simDF)
        return Evaluator(self.subPeriod(self.startDate+timedelta(days=1), self.endDate), simulations, color)

    def subPeriod(self, startDate, endDate):
        """ Build a new Observations object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Observations object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError("startDate must be >= {} !!!".format(self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError("endDate must be <= {} !!!".format(self.endDate))

        # select sub-period
        idx = (self.obsDF.index.date >= startDate) & (self.obsDF.index.date < endDate + timedelta(days=self.forecastHorizon))
        obsDF = self.obsDF.loc[idx]

        return self.__class__(species=self.species, start=startDate, end=endDate,
            stations=self.stations, seriesType=self.seriesType, obsDF=obsDF,
            forecastHorizon=self.forecastHorizon)

    def simVSobs(self, grid, time, point_size=20, vmin=None, vmax=None,
        cmap=None, colors=None, bounds=None, outputFile=None, file_formats=['png']):
        """ Scatter plot of observations above simulation raster. 

        Parameters
        ----------
        grid : evaltools.interpolation.Grid
            Grid object that must contain data for the species of
            the current Observations object.
        time : datetime.datetime or datetime.date
            Time for which to plot the observation values. This time
            must be contained in the current Observations object.
        point_size : float
            Point size (as define in matplotlib.pyplot.scatter).
        vmin, vmax : None or scalar
            Min and max values for the legend colorbar. If None, these
            values are found automatically. 
        cmap : None or matplotlib.colors.Colormap object
            Colors used for plotting (default: matplotlib.cm.jet).
        colors : None or list of str
            List of color used in the chart if you want to discretize
            the values.
        bounds : None or list of scalar
            Boundary values for each category if you want to discretize
            the values. Arguments vmin and vmax must not be None, and the
            boundary values contained between vmin and vmax.
            Ignored if colors is None.
        outputFile : None str
            File where to save the plot (without extension).
            If None, the figure is shown in a popping window.
        file_formats : list of str
            List of file extensions.

        """
        import matplotlib.pyplot as plt
        from matplotlib import cm
        from mpl_toolkits.basemap import Basemap
        import matplotlib as mpl

        # define default colormap
        if cmap is None:
            cmap = cm.jet

        # plotting
        plt.close('all')
        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111)

        # draw coast lines with Basemap
        m = Basemap(resolution='i', projection='merc',
            llcrnrlat=grid.minLat, urcrnrlat=grid.maxLat,
            llcrnrlon=grid.minLon, urcrnrlon=grid.maxLon,
            lat_ts=(grid.minLat + grid.maxLat)/2)
        m.drawcoastlines(linewidth=0.5, color='k', zorder=1)

        # discretizing values if colors is not None
        norm = None
        if colors is not None:
            cmap = mpl.colors.ListedColormap(colors)
            if bounds is not None:
                bounds = [vmin] + bounds + [vmax]
                norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        # plot simulation raster
        m.imshow(grid.grids[self.species],
                   extent=(grid.minLon, grid.maxLon, grid.minLat, grid.maxLat),
                   zorder=0,
                   cmap=cmap,
                   vmin=vmin,
                   vmax=vmax,
                   norm=norm
                   )

        # plot observation values
        m.scatter(list(self.stations['lon']), list(self.stations['lat']),
            c=self.obsDF.loc[time],
            latlon=True,
            edgecolors='k',
            linewidths=1,
            s=point_size,
            zorder=2,
            cmap=cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax
            )
        cb = m.colorbar(ticks=bounds)

        # save figure
        if outputFile is None:
            plt.show()
        else:
            for ext in file_formats:
                plt.savefig((outputFile+".{ext}").format(ext=ext))

    def dailyMean(self, availability_ratio=0.75):
        """ Build Observations object working on daily mean.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily mean.

        Returns
        -------
            Observations object with seriesType = 'daily' and with data
            corresponding to the computed daily mean.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        obsDF = evt.timeseries.daily_mean(self.obsDF, availability_ratio)

        return self.__class__(self.species, self.startDate, self.endDate, self.stations,
            obsDF=obsDF, forecastHorizon=self.forecastHorizon, seriesType='daily')

    def dailyMax(self, availability_ratio=0.75):
        """ Return Observations object working on daily maximum.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily maximum.

        Returns
        -------
            Observations object with seriesType = 'daily' and with data
            corresponding to the computed daily maximum.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        obsDF = evt.timeseries.daily_max(self.obsDF, availability_ratio=availability_ratio)

        return self.__class__(self.species, self.startDate, self.endDate, self.stations,
            obsDF=obsDF, forecastHorizon=self.forecastHorizon, seriesType='daily')

    def movingAverageDailyMax(self, availability_ratio=0.75, keep_first_day=True):
        """ Return Observations object working on the daily maximum of the moving average.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available to compute the average for a given 8-hour window,
            and also minimal rate of values available for a given day to compute the daily
            maximum of the moving average. For example with availability_ratio=0.75, a daily maximum
            eight hours average can only be calculated if 18 eight hours average
            are available each of which requires 6 hourly values to be available.
        keep_first_day : bool
            If True, the first day of the period is kept even though
            all of its values are equal to np.nan. If false, the first
            day of the period is dropped.

        Returns
        -------
            Observations object with seriesType = 'daily' and with data corresponding to the
            computed daily maximum of the moving average.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        obsDF = evt.timeseries.moving_average_daily_max(self.obsDF, availability_ratio)

        # keep first day by setting it to np.nan
        if keep_first_day is True:
            obsDF.loc[obsDF.index[0]-timedelta(days=1)] = np.nan
            obsDF.sort_index(inplace=True)
            startDate = self.startDate
        else:
            startDate = self.startDate+timedelta(days=1)

        return self.__class__(self.species, startDate, self.endDate, self.stations,
            obsDF=obsDF, forecastHorizon=self.forecastHorizon, seriesType='daily')

    def filteredSeries(self, availability_ratio=0.75):
        """ Return Observations object working on the filtered series.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available for a given hour to compute the daily
            filtered series for this hour in all days.

        Returns
        -------
            Observations object with seriesType = 'hourly' and with data
            corresponding to the computed filtered series.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        obsDF = evt.timeseries.filtered_series(self.obsDF, availability_ratio)

        return self.__class__(self.species, self.startDate, self.endDate, self.stations,
            obsDF=obsDF, forecastHorizon=self.forecastHorizon, seriesType='hourly')

    def normalized_series(self):
        """ Return Observations object with normalized series.

        This method normalizes each series of observations by substracting
        its median and dividing by its interquartile range.

        Returns
        -------
            Observations object with seriesType = 'hourly' and with data
            corresponding to the computed normalized series.

        """
        obsDF = evt.timeseries.normalized_series(self.obsDF)
        return self.__class__(self.species, self.startDate, self.endDate, self.stations,
            obsDF=obsDF, forecastHorizon=self.forecastHorizon, seriesType=self.seriesType)


class Simulations(object):
    """ Class gathering simulations of the studied case.

    An object of class Simulations will be specific to one species,
    one series type, one period and one model.

    Parameters
    ----------
    species : str
        Species name (ex: "o3").
    model : str
        Name of the model that produced the simulated data.
    start : datetime.date
        Start day of the studied period.
    end : datetime.date
        End day (included) of the studied period.
    seriesType : str
        It can be 'hourly' or 'daily'.
    simDF : list of pandas.DataFrame
        List of DataFrame with datetime index (shifted from the studied period according
        to the forcast day it represents) and containing one column per station.
        Each element of the list correspond to a forecast day. For example, if 'forecastHorizon'
        is 4, the list will be of length 4.

    """
    def __init__(self, species, model, start, end, seriesType, simDF):
        self.species = species
        if type(start) is date:
            self.startDate = start
            self.endDate = end
        else:
            self.startDate = date(int(start[0:4]), int(start[4:6]), int(start[6:8]))
            self.endDate = date(int(end[0:4]), int(end[4:6]), int(end[6:8]))
        self.forecastHorizon = len(simDF)

        # initialize simDF
        self.simDF = simDF
        for df in self.simDF:
            df.index = pd.to_datetime(df.index)
        self.model = model
        self.seriesType = seriesType

        # check if start/end well correspond to simDF index
        nb_days = (self.endDate-self.startDate).days + 1
        if seriesType == 'hourly':
            start_time = datetime.combine(self.startDate,time(0))
            steps = pd.DatetimeIndex([start_time+timedelta(hours=h) for h in range(nb_days*24)])
        elif seriesType == 'daily':
            start_time = self.startDate
            steps = pd.DatetimeIndex([start_time+timedelta(days=d) for d in range(nb_days)])
        else:
            raise evt.EvaltoolsError("seriesType argument must be either 'hourly' or 'daily' !!!")
        try:
            for d, df in enumerate(self.simDF):
                assert((steps+timedelta(days=d) == df.index).all())
        except:
            raise evt.EvaltoolsError("DataFrames index of simDF do not fit start/end arguments !!!")

    @classmethod
    def fromTimeSeries(cls, generic_file_path, stationsIdx, species, model,
        start, end, forecastHorizon=1, correc_unit=1, seriesType='hourly'):
        """ Class method used to construct an object from timeseries files.

        Parameters
        ----------
        generic_file_path : str
            Generic path of timeseriesfiles with {year} instead of
            the year number, {station} instead of the station name and {forecastDay}
            instead of the forecast day number.
        stationsIdx : list of str
            List of the names of studied stations.
        species : str
            Species name (ex: "o3").
        model : str
            Name of the model that produced the simulated data.
        start : datetime.date
            Start day of the studied period.
        end : datetime.date
            End day (included) of the studied period.
        forecastHorizon : int
            Number of day corresponding to the forcast horizon of the model.
        correc_unit : float
            Multiplicative factor applied to original values.
        seriesType : str
            It can be 'hourly' or 'daily'.

        """
        simDF = [evt.timeseries.get_DF(stationsIdx=stationsIdx,
                     generic_file_path=generic_file_path.format(year="{year}", station="{station}", forecastDay=d),
                     start_date=start,
                     end_date=end,
                     lag=d,
                     seriesType=seriesType)
                 for d in range(forecastHorizon)]
        obj = cls(species=species, model=model, start=start, end=end,
            seriesType=seriesType, simDF=simDF)
        return obj

    @classmethod
    def fromEvtDataFrame(cls, model, edf_list, listing_path, forecastHorizon=1, correc_unit=1):
        """ Class method used to construct an object from evaltools.dataframe.EvtDataFrame objects.

        Stations kept in the returned EvtDataFrame are the intersection
        of stations from the EvtDataFrame objects of edf_list.

        Parameters
        ----------
        model : str
            Name of the model.
        edf_list : list of evaltools.dataframe.EvtDataFrame objects
            EvtDataFrame objects corresponding to the different
            forecast terms. The order in the list matters.
        listing_path : str
            Path of the station listing. This listing only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correct_unit : float
            Factor to apply to original values.

        Returns
        -------
            evaltools.evaluator.Simulations object

        """
        # check forecastHorizon
        if len(edf_list) != forecastHorizon:
            raise evt.EvaltoolsError("Length of edf_list argument must be equal to forecastHorizon argument !!!")

        # define station list as the intersection of stations from EvtDataFrame objects
        sim_stations = reduce(np.intersect1d, [edf.DF.columns for edf in edf_list])
        stations = evt.utils.read_listing(listing_path)
        if not np.isin(sim_stations, stations.index).all():
            raise evt.EvaltoolsError("Listing does not contain all stations of the EvtDataFrame objects !!!")
        stations = stations.loc[sim_stations]

        # check metadata
        species = edf_list[0].species
        startDate = edf_list[0].startDate
        endDate = edf_list[0].endDate
        for fd, edf in enumerate(edf_list):
            if species != edf.species or edf.startDate != (startDate+timedelta(days=fd)) or edf.endDate != (endDate+timedelta(days=fd)):
                raise evt.EvaltoolsError("Metadata of EvtDataFrame objects from edf_list are not compatible !!!")

        # build Simulations object
        simDF = [edf.DF[stations.index] for edf in edf_list]
        if correc_unit != 1:
            simDF = [df*correc_unit for df in simDF]
        obj = cls(species=species, model=model, start=startDate, end=endDate, seriesType='hourly', simDF=simDF)

        return obj

    @classmethod
    def fromHDF(cls, startDate, endDate, species, model, stations, hdf_path, listing_path, forecastHorizon=1, correc_unit=1):
        """ Class method used to construct an object from HDF5 files.

        Parameters
        ----------
        startDate : datetime.date
            Starting day of the studied period.
        endDate : datetime.date
            Ending day (included) of the studied period.
        species : str
            Species name (must be a key of the hdf file).
        model : str
            Name of the model.
        stations : 1D array of str
            List of stations to keep in the resulting object.
        hdf_path : str
            Path of the input file. The file path must contain
            '{forecastDay}' instead of the forecast day of the file.
            Input data can be splitted over two hdf files
            corresponding to two different years. In that case, the
            file path must contain '{year}'.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.

        Returns
        -------
            evaltools.evaluator.Simulations object

        """
        sim_evtDF_list = []
        for fd in range(forecastHorizon):
            store_list = [evt.dataframe.Store(hdf_path.format(year=year, forecastDay=fd), read_only=True)
                          for year in np.unique([startDate.year, endDate.year])]
            edf_list = [store.getDF(species=species, dsetName=model, startDate=startDate+timedelta(days=fd), endDate=endDate+timedelta(days=fd)) for store in store_list]
            df = evt.dataframe.EvtDataFrame(
                stations=np.intersect1d(reduce(np.union1d, [edf.DF.columns for edf in edf_list]), stations),
                startDate=startDate+timedelta(days=fd),
                endDate=endDate+timedelta(days=fd),
                species=species)
            for edf in edf_list:
                df.updateFromEvtDF(edf)
            sim_evtDF_list.append(df)

        simulations = evt.evaluator.Simulations.fromEvtDataFrame(
            model=model,
            edf_list=sim_evtDF_list,
            listing_path=listing_path,
            forecastHorizon=forecastHorizon,
            correc_unit=correc_unit)

        return simulations

    @classmethod
    def from_csv(cls, inputFiles, seriesType, species, model, start, end, forecastHorizon=4):
        """ Class method used to construct an object from a csv file.

        Parameters
        ----------
        inputFiles : list of str
            Paths of csv file corresponding to Simulations.simDF attribute.
        seriesType : str
            It can be 'hourly' or 'daily'.
        species : str
            Species name (ex: "o3").
        model : str
            Name of the model that produced the simulated data.
        start : str or datetime.date
            Start day of the studied period. If str, it must
            be of the form "yyyymmdd".
        end : str or datetime.date
            End day (included) of the studied period. If str, it must
            be of the form "yyyymmdd".
        forecastHorizon : int
            Number of days corresponding to the forcast horizon of the model.

        """
        for f in inputFiles:
            if not isfile(f):
                raise evt.EvaltoolsError("Missing file : {0} !!!".format(f))
        obj = cls(species, model, start, end, seriesType, forecastHorizon=forecastHorizon)
        obj.simDF = [pd.read_csv(f, sep=' ', index_col=0) for f in inputFiles]
        for fd in range(obj.forecastHorizon):
            obj.simDF[fd].index = pd.to_datetime(obj.simDF[fd].index, format='%Y%m%d%H')
        return obj


    def to_csv(self, filePath):
        """ Save timeseries dataframe as csv.

        Parameters
        ----------
        filePath : str
            File path with {forecastDay} instead of forecast day number.

        """
        for forecast_day in range(self.forecastHorizon):
            self.simDF[forecast_day].to_csv(filePath.format(forecastDay=forecast_day),
                                             sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')

    def subPeriod(self, startDate, endDate):
        """ Build a new Simulations object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Simulations object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError("startDate must be >= {} !!!".format(self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError("endDate must be >= {} !!!".format(self.endDate))

        # select sub-period
        simDF = []
        for fday in range(self.forecastHorizon):
            idx = (self.simDF[fday].index.date >= startDate + timedelta(days=fday)) & (self.simDF[fday].index.date <= endDate + timedelta(days=fday))
            simDF.append(self.simDF[fday].loc[idx])

        return self.__class__(species=self.species, model=self.model,
            start=startDate, end=endDate, seriesType=self.seriesType,
            simDF=simDF)

    def check_values(self, threshold, drop=False, filePath=None):
        """ Check if observation values exceed a threshold.

        If there are values above the threshold, a message is printed
        and these values are set to nan if drop == True.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        drop : bool
            If True, values above the threshold are set to nan.
        filePath : None or str
            File path where to save the names of stations that exceed
            the threshold.

        """
        for fd in range(self.forecastHorizon):
            idx = (self.simDF[fd] > threshold)
            nb_val_above = np.sum(idx.values)
            if nb_val_above > 0:
                print("*** {nb} values above threshold for the forecast day term {fd} in simulation {model} !!! values are {action}.".format(nb=nb_val_above, fd=fd, model=self.model, action='kept'*(not drop)+'set to np.nan'*drop))
                if filePath is not None:
                    self.simDF[fd].stack().loc[idx.stack()].to_csv(filePath.format(forecastDay=fd), sep=' ', na_rep='nan', float_format='%g', header=False, index=True, date_format='%Y%m%d%H')
                if drop is True:
                    self.simDF[fd][idx] = np.nan
            else:
                print("0 value above threshold in simulation {model}".format(model=self.model))

    def dailyMean(self, availability_ratio=0.75):
        """ Build Simulations object working on daily mean.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily mean.

        Returns
        -------
            Simulations object with seriesType = 'daily' and with data corresponding to the
            computed daily mean.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        simDF = [evt.timeseries.daily_mean(self.simDF[day], availability_ratio) for day in range(self.forecastHorizon)]

        return self.__class__(self.species, self.model, self.startDate, self.endDate,
            simDF=simDF, seriesType='daily')

    def dailyMax(self, availability_ratio=0.75):
        """ Return Simulations object working on daily maximum.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily maximum.

        Returns
        -------
            Simulations object with seriesType = 'daily' and with data corresponding to the
            computed daily maximum.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        simDF = [evt.timeseries.daily_max(self.simDF[day], availability_ratio=availability_ratio) for day in range(self.forecastHorizon)]

        return self.__class__(self.species, self.model, self.startDate, self.endDate,
            simDF=simDF, seriesType='daily')

    def movingAverageDailyMax(self, availability_ratio=0.75, keep_first_day=True):
        """ Return Simulations object working on the daily maximum of the moving average.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available to compute the average for a given 8-hour window,
            and also minimal rate of values available for a given day to compute the daily
            maximum of the moving average. For example with availability_ratio=0.75, a daily maximum
            eight hours average can only be calculated if 18 eight hours average
            are available each of which requires 6 hourly values to be available.
        keep_first_day : bool
            If True, the first day of the period is kept even though
            all of its values are equal to np.nan. If false, the first
            day of the period is dropped.

        Returns
        -------
            Simulations object with seriesType = 'daily' and with data corresponding to the
            computed daily maximum of the moving average.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        simDF = [evt.timeseries.moving_average_daily_max(self.simDF[day], availability_ratio) for day in range(self.forecastHorizon)]

        # keep first day by setting it to np.nan
        if keep_first_day is True:
            for fd in range(self.forecastHorizon):
                simDF[fd].loc[simDF[fd].index[0]-timedelta(days=1)] = np.nan
                simDF[fd].sort_index(inplace=True)
            startDate = self.startDate
        else:
            startDate = self.startDate+timedelta(days=1)

        return self.__class__(self.species, self.model, startDate, self.endDate,
            simDF=simDF, seriesType='daily')

    def filteredSeries(self, availability_ratio=0.75):
        """ Return Simulations object working on the filtered series.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available for a given hour to compute the daily
            filtered series for this hour in all days.

        Returns
        -------
            Simulations object with seriesType = 'hourly' and with data
            corresponding to the computed filtered series.

        """
        if self.seriesType not in ['hourly']:
            print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
            return None

        simDF = [evt.timeseries.filtered_series(self.simDF[day], availability_ratio) for day in range(self.forecastHorizon)]

        return self.__class__(self.species, self.model, self.startDate, self.endDate,
            simDF=simDF, seriesType='hourly')

    def normalized_series(self):
        """ Return Simulations object with normalized series.

        This method normalizes each series of simulations by substracting
        its median and dividing by its interquartile range.

        Returns
        -------
            Simulations object with seriesType = 'hourly' and with data
            corresponding to the computed normalized series.

        """
        simDF = [evt.timeseries.normalized_series(self.simDF[day]) for day in range(self.forecastHorizon)]
        return self.__class__(self.species, self.model, self.startDate, self.endDate,
            simDF=simDF, seriesType=self.seriesType)


def load(inputFilePath):
    """ evaluator.Evaluator object loader

    Load an evaluator.Evaluator object saved in binary format
    with evaluator.Evaluator.dump method.

    Parameters
    ----------
    inputFilePath : str
        Path of the binary file to load.

    Returns
    -------
        evaluator.Evaluator object

    """
    with open(inputFilePath, 'rb') as f:
        obj = pickle.load(f)
    if type(obj) == evt.evaluator.Evaluator:
        return obj
    else:
        raise evt.EvaltoolsError("*** {} not an evaluator.Evaluator object !!!".format(inputFilePath))


class Evaluator(object):
    """ Class gathering observations and simulations of the studied case.

    An object of class :class:`~evaltools.evaluator.Simulations` will be specific to one species,
    one series type, one period and one model. This class contain several methods
    to compare simulations and observations.

    Parameters
    ----------
    observations : evaluator.Observations object
        Observations used to construct the Evaluator object.
    simulations : evaluator.Simulations object
        Simulations used to construct the Evaluator object.
    color : str
        Default color that will be used in plotting functions.

    """
    def __init__(self, observations, simulations, color='k'):
        if all([df.empty for df in simulations.simDF]):
            raise evt.EvaltoolsError("Trying to construct Evaluator object with empty simulations !!!")
        if observations.obsDF.empty:
            raise evt.EvaltoolsError("Trying to construct Evaluator object with empty observations !!!")
        try:
            assert(observations.species == simulations.species)
            assert(observations.startDate == simulations.startDate)
            assert(observations.endDate == simulations.endDate)
            assert(observations.forecastHorizon >= simulations.forecastHorizon)
            assert(observations.seriesType == simulations.seriesType)
        except:
            raise evt.EvaltoolsError("observations and simulations do not match !!!")

        self.__dict__.update(observations.__dict__)
        self.__dict__.update(simulations.__dict__)

        # only keep stations shared by obs and sim
        sim_stations = np.array([])
        for i in range(self.forecastHorizon):
            sim_stations = np.union1d(sim_stations, np.array(self.simDF[i].columns))
        self.stations = self.stations.loc[np.intersect1d(sim_stations, np.array(self.obsDF.columns))]
        self.obsDF = self.obsDF[self.stations.index]

        # give all forecast days simulations DF the same station list (at the cost of introducing nan)
        for i in range(self.forecastHorizon):
            df_temp = pd.DataFrame(np.nan, index=self.simDF[i].index, columns=self.stations.index)
            df_temp.update(self.simDF[i])
            self.simDF[i] = df_temp

        # color attribute used in plotting function
        self.color = color

        # dictionary where to save score results
        self.results = {}

    @classmethod
    def fromHDF(cls, startDate, endDate, species, model, stations, obs_hdf_path,
        sim_hdf_path, listing_path, forecastHorizon=1, correc_unit=1,
        color='k'):
        """ Class method used to construct an object from HDF5 files.

        The architecture of HDF5 files must be the one used by
        :class:`~evaltools.dataframe.Store`.

        Parameters
        ----------
        startDate : datetime.date
            Starting day of the studied period.
        endDate : datetime.date
            Ending day (included) of the studied period.
        species : str
            Species name (must be a key of the hdf file).
        model : str
            Name of the model.
        stations : 1D array of str
            List of stations to keep in the resulting object.
        obs_hdf_path : str
            Path of the input file. Input data can be splitted over
            two hdf files corresponding to two different years. In
            that case, the file path must contain '{year}'.
        sim_hdf_path : str
            Path of the input file. The file path must contain
            '{forecastDay}' instead of the forecast day of the file.
            Input data can be splitted over two hdf files
            corresponding to two different years. In that case, the
            file path must contain '{year}'.
        listing_path : str
            Path of the station listing. This listing is only used
            to get metadata of the stations.
        forecastHorizon : int
            Number of forecasted days.
        correc_unit : float
            Factor to apply to original values.
        color : str
            Default color that will be used in plotting functions.

        Returns
        -------
            evaltools.evaluator.Evaluator object

        """
        observations = Observations.fromHDF(startDate, endDate, species,
            obs_hdf_path, listing_path, forecastHorizon=forecastHorizon, correc_unit=correc_unit)
        simulations = Simulations.fromHDF(startDate, endDate, species,
            model, stations, sim_hdf_path, listing_path, forecastHorizon=forecastHorizon, correc_unit=correc_unit)
        return cls(observations, simulations, color=color)

    def dump(self, outputFilePath):
        """ evaluator.Evaluator object loader

        Save an evaluator.Evaluator object in binary format.
        Use evaltools.evaluator.load function to get the object back.

        Parameters
        ----------
        outputFilePath : str
            Path of the output binary file.

        """
        with open(outputFilePath, 'wb') as f:
            pickle.dump(self, f)

    def summary(self):
        """ Print a summary of the object. """
        print("Model : {0}".format(self.model))
        print("Species : {0}".format(self.species))
        print("Period : {0} - {1}".format(self.startDate.strftime('%Y%m%d'), self.endDate.strftime('%Y%m%d')))
        print("Forecast horizon : {0}".format(self.forecastHorizon))

    def subPeriod(self, startDate, endDate):
        """ Build a new Evaluator object define on a shorter period.

        Parameters
        ----------
        startDate : datetime.date
            Starting date of the new object.
        endDate : datetime.date
            Ending date of the new object.

        Returns
        -------
            Evaluator object.

        """
        # check args integrity
        if startDate < self.startDate:
            raise evt.EvaltoolsError("startDate must be >= {} !!!".format(self.startDate))
        if endDate > self.endDate:
            raise evt.EvaltoolsError("endDate must be <= {} !!!".format(self.endDate))

        # recreate Observations and Simulations
        obs = Observations(species=self.species, start=self.startDate, end=self.endDate,
            stations=self.stations, seriesType=self.seriesType, obsDF=self.obsDF,
            forecastHorizon=self.forecastHorizon)
        sim = Simulations(species=self.species, model=self.model,
            start=self.startDate, end=self.endDate, seriesType=self.seriesType,
            simDF=self.simDF)

        # take sub period
        observations = obs.subPeriod(startDate, endDate)
        simulations = sim.subPeriod(startDate, endDate)

        return self.__class__(observations, simulations, color=self.color)

    def stationSubList(self, station_list, inplace=True):
        """ Drop stations not contained in the given list.

        Parameters
        ----------
        station_list : 1D array of str.
            List of stations to keep.
        inplace : bool
            If True, the Evaluator object is modified inplace,
            else, a new object is returned.

        """
        if inplace is True:
            result = self
        else:
            result = copy.deepcopy(self)

        # check if station_list is included in self.stations.index
        stations = np.intersect1d(self.stations.index, station_list)
        missing_stations =  np.array(station_list)[~np.isin(station_list, stations)]
        if missing_stations.shape[0] > 0:
            print("Warning: stations {} from station_list were not found in the original object !!!".format(missing_stations))

        # drop unsought stations
        result.stations = result.stations.loc[stations]
        result.obsDF = result.obsDF[stations]
        for fd in range(result.forecastHorizon):
            result.simDF[fd] = result.simDF[fd][stations]

        if inplace is not True:
            return result

    def subArea(self, minLon, maxLon, minLat, maxLat, inplace=True):
        """ Drop stations not contained within the given lat/lon boundaries.

        Parameters
        ----------
        minLon, maxLon, minLat, maxLat : scalars
            Lat/lon boundaries.
        inplace : bool
            If True, the Evaluator object is modified inplace,
            else, a new object is returned.

        """
        kept_stations_idx = reduce(np.logical_and,
            [self.stations['lat'] >= minLat, self.stations['lat'] <= maxLat,
            self.stations['lon'] >= minLon, self.stations['lon'] <= maxLon])
        if inplace is True:
            self.stationSubList(self.stations.index[kept_stations_idx],
                inplace=inplace)
        else:
            return self.stationSubList(self.stations.index[kept_stations_idx],
                inplace=inplace)

    def time_series_to_csv(self, obsOutputFile, simOutputFile):
        """ Save timeseries dataframes as csv.

        Parameters
        ----------
        obsOutputFile : str
            File path where to save observed timeseries.
        simOutputFile : str
            File path where to save simulated timeseries. The path name
            must contain {forecastDay} instead of the forecast day number.

        """
        self.obsDF.to_csv(obsOutputFile, sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')
        for forecast_day in range(self.forecastHorizon):
            self.simDF[forecast_day].to_csv(simOutputFile.format(forecastDay=forecast_day), sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')

    def get_obs(self, forecast_day, start_end=None):
        """ Get observations according to the forecast day.

        Parameters
        ----------
        forecast_day : int
            Forecast day for which to get observations.
        start_end : None or list of two datetime.date objects
            Dates between which getting the data.

        Returns
        -------
            pandas.DataFrame of observations on time period corresponding to the forecast day.

        """
        if start_end is not None:
            start = start_end[0] + timedelta(days=forecast_day)
            end = start_end[1] + timedelta(days=forecast_day)
        else:
            start = self.startDate + timedelta(days=forecast_day)
            end = self.endDate + timedelta(days=forecast_day)

        idx = (self.obsDF.index.date >= start) & (self.obsDF.index.date <= end)

        return self.obsDF.loc[idx]

    def reconstruct_observations(self):
        """ Get Observations object from the Evaluator object.

        Returns
        -------
            evaltools.evaluator.Observations object.

        """
        return Observations(self.species, self.startDate, self.endDate,
            self.stations, obsDF=self.obsDF,
            forecastHorizon=self.forecastHorizon, seriesType=self.seriesType)

    def reconstruct_simulations(self):
        """ Get Simulations object from the Evaluator object.

        Returns
        -------
            evaltools.evaluator.Simulations object.

        """
        return Simulations(self.species, self.model,
            self.startDate, self.endDate, simDF=self. simDF,
            seriesType=self.seriesType)

    def get_sim(self, forecast_day, start_end=None):
        """ Get simulations according to the forecast day.

        Parameters
        ----------
        forecast_day : int
            Forecast day for which to get simulations.
        start_end : None or list of two datetime.date objects
            Dates between which getting the data.

        Returns
        -------
            pandas.DataFrame of simulations corresponding to the forecast day.

        """
        if start_end is not None:
            start = start_end[0] + timedelta(days=forecast_day)
            end = start_end[1] + timedelta(days=forecast_day)

            idx = (self.simDF[forecast_day].index.date >= start) & (self.simDF[forecast_day].index.date <= end)

            return self.simDF[forecast_day].loc[idx]

        return self.simDF[forecast_day]

    def stationScores(self, score_list, outputFile=None, availability_ratio=0.75):
        """ Compute scores per station for each forecast day.

        This method may not be appropriate for hourly time steps data.
        You can also checkout FTscores method.

        Parameters
        ----------
        score_list : list of str
            List of computed scores.
        outputFile : str
            File where to save the result. The file name must contain {forecastDay}
            instead of the forecast day number. If None, result is not saved in csv.
        availability_ratio : float
            Minimal rate of data available on the period required per forecast
            day to compute the scores.

        Returns
        -------
        res : dict
            Dictionary with one key per forecast day ('D0', 'D1', ...) containing pandas.DataFrames with
            station names as index and one column per computed score.

        """
        # if self.seriesType not in ['daily']:
        #     print("*** Object's attribute 'seriesType' must be 'daily' to use this method !!!")
        #     return None

        res = {}
        for forecast_day in range(self.forecastHorizon):
            obs = self.get_obs(forecast_day)
            sim = self.get_sim(forecast_day)
            statsDFsta = evt.scores.stats2D(obs, sim, score_list, axis=0, threshold=float(availability_ratio), keep_nan=True)
            if outputFile is not None:
                statsDFsta.to_csv(outputFile.format(forecastDay=forecast_day), sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')
            res["D{}".format(str(forecast_day))] = statsDFsta
        return res

    def quarterlyMedianScore(self, filePath, score='RMSE', forecastDay=0, availability_ratio=0.75):
        """ Compute median on station scores.
        
        Parameters
        ----------
        filePath : str
            File where to find previous values and save the result.
        score : str
            The score to process.
        forecastDay : int
            Integer corresponding to the chosen forecastDay used for computation.
        availability_ratio : float
            Minimal rate of data available on the period required per forecast
            day to compute the scores.

        Returns
        -------
            The median for the computed score.

        """
        stats = self.stationScores(score_list=[score], availability_ratio=availability_ratio)['D{}'.format(forecastDay)][score]

        if isfile(filePath):
            df = pd.read_csv(filePath, sep=' ', index_col=0)
        else:
            df = pd.DataFrame(columns=[score])

        quarter = evt.quarter.Quarter(self.startDate, self.endDate)
        idx = quarter.__repr__()

        res = np.nanmedian(stats)
        df.loc[idx] = res
        df.to_csv(filePath, sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')
        return res

    def timeScores(self, score_list, outputFile=None, min_nb_sta=10):
        """ Compute scores per time.

        Compatible both for hourly and daily series types.

        Parameters
        ----------
        score_list : list of str
            List of computed scores.
        outputFile : str. File where to save the result. The file name must contain {forecastDay}
            instead of the forecast day number. If None, result is not saved in csv.
        min_nb_sta : int
            Minimal number of station values available in both obs and sim required
            per datetime to compute the scores.

        Returns
        -------
            Dictionary with one key per forecast day ('D0', 'D1', ...) containing pandas.DataFrames with datetime index and one column per computed score.

        """
        try:
            assert self.results['timeScores']['args'] == {'score_list': score_list, 'min_nb_sta': min_nb_sta}
            res = self.results['timeScores']['result']

        except:
            # print("computing timescores")
            res = {}
            for fd in range(self.forecastHorizon):
                obs = self.get_obs(fd)
                sim = self.get_sim(fd)
                statsDFsta = evt.scores.stats2D(obs, sim, score_list, axis=1, threshold=int(min_nb_sta))
                if outputFile is not None:
                    statsDFsta.to_csv(outputFile.format(forecastDay=fd), sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')
                res["D{}".format(str(fd))] = statsDFsta
            self.results['timeScores'] = {'args': {'score_list': score_list, 'min_nb_sta': min_nb_sta}, 'result': res}

        return res

    def meanTimeScores(self, score_list, outputFile=None, min_nb_sta=10, availability_ratio=0.75):
        """ Compute the mean of time scores for each forecast time.

        Parameters
        ----------
        score_list : list of str
            List of computed scores.
        outputFile : str
            File where to save the result. The file name must contain {forecastDay}
            instead of the forecast day number. If None, result is not saved in csv.
        min_nb_sta : int
            Minimal number of station values available in both obs and sim required
            per datetime to compute the scores (before applying mean for each
            forecast time).
        availability_ratio : float
            Minimal rate of data (computed scores per time) available on the period required per forecast
            time to compute the mean scores.

        Returns
        -------
            Dictionary, with one key per score, that contains lists of the means for each forecast time. For example if the forecast horizon is 4, the lists will be of length 96.

        """
        if self.seriesType not in ['hourly']:
            raise evt.EvaltoolsError("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")

        # scores computing
        stats = self.timeScores(score_list, min_nb_sta=min_nb_sta)

        res = {}
        nb_days = (self.endDate - self.startDate).days + 1
        for sc in score_list:
            Y = []
            for fd in range(self.forecastHorizon):
                matrix = stats["D{}".format(str(fd))][sc].values.reshape(nb_days, 24)
                idx = np.sum(np.isnan(matrix), axis=0) >= (availability_ratio*matrix.shape[0])
                with np.warnings.catch_warnings():
                    np.warnings.filterwarnings('ignore', 'Mean of empty slice')
                    means = np.nanmean(matrix, axis=0)
                means[idx] = np.nan
                Y.extend(list(means))
            res[sc] = Y
        if outputFile is not None:
            pd.DataFrame(res).to_csv(outputFile, sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')

        return res

    def FTscores(self, score_list, availability_ratio=0.75, outputFile=None):
        """ Compute forecast time scores for each station.

        Only available for hourly time step data.

        Parameters
        ----------
        score_list : list of str
            List of scores to compute.
        availability_ratio : float
            Minimal rate of data available on the period required per forecast
            time to compute the scores for each station.
        outputFile : str
            File where to save the result. If None, result is not saved in csv.
            The file name must contain {score} instead of the score name. A file in created
            for each score.

        Returns
        -------
            Dictionary, with one key per score, that contains pandas.DataFrame with one row per station and one column per forecast time except the first two columns that are latitude and longitude.

        """
        try:
            assert self.results['FTscores']['args'] == {'score_list': score_list, 'availability_ratio': availability_ratio}
            res = self.results['FTscores']['result']

        except:
            if self.seriesType not in ['hourly']:
                print("Object's attribute 'seriesType' must be 'hourly' to use this method !!!")
                return None
            res = {}
            for sc in score_list:
                res[sc] = self.stations[['lat', 'lon']]
                res[sc].columns = pd.MultiIndex.from_product([['coord'], ['lat', 'lon']])
            for day in range(self.forecastHorizon):
                obs = self.get_obs(day)
                sim = self.get_sim(day)
                dailyStats = evt.scores.FTstats(obs, sim, score_list, availability_ratio, day)
                midx = pd.MultiIndex.from_product([['D{}'.format(day)],range(24)])
                for sc in score_list:
                    dailyStats[sc].columns = midx
                    res[sc] = pd.concat([res[sc], dailyStats[sc]], axis=1, sort=False)
            if outputFile is not None:
                for sc in score_list:
                    file_path = outputFile.format(score=sc)
                    res[sc].insert(0, ('station','station'), res[sc].index)
                    res[sc].to_csv(file_path, sep=',', na_rep='nan', float_format='%g', header=True, index=False, date_format='%Y%m%d%H')
            self.results['FTscores'] = {'args': {'score_list': score_list, 'availability_ratio': availability_ratio}, 'result': res}

        return res

    def medianStationScores(self, score_list, availability_ratio=0.75, min_nb_sta=10, outputFile=None):
        """ Compute median station scores for each forecast time.

        Only available for hourly time step data.

        Parameters
        ----------
        score_list : list of str
            List of scores to compute.
        availability_ratio : float
            Minimal rate of data available on the period required per forecast
            time to compute the scores for each station.
        min_nb_sta : int
            Minimal number of station values available in both obs and sim required
            to compute the median.
        outputFile : str
            File where to save the result. If None, result is not saved in csv.
            The file name must contain {score} instead of the score name. A file in created
            for each score.

        Returns
        -------
            pandas.DataFrame with one row per forecast time and one column per score.

        """
        df_dict = self.FTscores(score_list, availability_ratio=availability_ratio, outputFile=None)

        med_scores = {}
        for sc in score_list:
            med_scores[sc] = np.nanmedian(df_dict[sc], axis=0)[2:]  # [2:] to remove lat lon

        res = pd.DataFrame(med_scores, index=range(self.forecastHorizon*24))

        idx_not_enough_values = np.sum(~list(df_dict.values())[0].isna(), axis=0)[2:].values < min_nb_sta
        res[idx_not_enough_values] = np.nan

        if outputFile is not None:
            pd.DataFrame(res).to_csv(outputFile, sep=' ', na_rep='nan', float_format='%g', header=True, index=True, date_format='%Y%m%d%H')

        return res

    def FDscores(self, score_list, forecast_days='all', threshold=0.75, outputFile=None):
        """ Compute forecast day scores for the model.

        Scores are computed using all data available for a given forecast day
        (no mean or median is taken).

        Parameters
        ----------
        score_list : list of str
            List of scores to compute.
        forecast_days : 'all' or list of int
            Forcast days used for computation. The returned DataFrame will contain one
            row per forcast day.
        threshold : int or float
            Minimal number (if type(threshold) is int) or minimal rate (if type(threshold) is float)
            of data available in both obs and sim required to compute the scores.
        outputFile : str
            File where to save the result. If None, result is not saved in csv.

        Returns
        -------
            pandas.DataFrame with one column per score and one row per forecast day.

        """
        res = pd.DataFrame()
        if forecast_days is 'all':
            forecast_days = range(self.forecastHorizon)
        for day in forecast_days:
            obs = self.get_obs(day).values.flatten()
            sim = self.get_sim(day).values.flatten()

            if type(threshold) is float:
                if threshold >= 1.:
                    print("*** Warning: threshold of type float >= 1")
                threshold = threshold*obs.shape[0]

            df = pd.DataFrame(evt.scores.stats2D_numpy(obs.reshape(obs.size, 1), sim.reshape(sim.size, 1), score_list, threshold=threshold), index=["D{}".format(day)], columns=score_list)
            res = res.append(df)

        if outputFile is not None:
            res.to_csv(file_path, sep=' ', na_rep='nan', float_format='%g', header=True, index=True)
        return res

    def dailyMean(self, availability_ratio=0.75):
        """ Build Evaluator object working on daily mean.

        This method computes the daily mean of observations and simulations
        of the current object.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily mean.

        Returns
        -------
            Evaluator object with seriesType = 'daily' and with data corresponding to the
            computed daily mean.

        """
        observations = self.reconstruct_observations()
        simulations = self.reconstruct_simulations()
        return self.__class__(observations.dailyMean(availability_ratio),
            simulations.dailyMean(availability_ratio), color=self.color)

    def dailyMax(self, availability_ratio=0.75):
        """ Return Evaluator object working on daily maximum.

        This method compute the daily maximum of observations and simulations
        of the current object and returns a new Evaluator object with
        seriesType = 'daily' and with data corresponding to the computed
        daily maxima.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available in a day required
            to compute the daily maximum.

        Returns
        -------
            Evaluator object with seriesType = 'daily' and with data corresponding to the
            computed daily maximum.

        """
        observations = self.reconstruct_observations()
        simulations = self.reconstruct_simulations()
        return self.__class__(observations.dailyMax(availability_ratio),
            simulations.dailyMax(availability_ratio), color=self.color)

    def movingAverageDailyMax(self, availability_ratio=0.75, keep_first_day=True):
        """ Return Evaluator object working on the daily maximum of the moving average.

        This method compute the daily maximum of the moving average
        for observations and simulations of the current object and returns
        a new Evaluator object.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available to compute the average for a given 8-hour window,
            and also minimal rate of values available for a given day to compute the daily
            maximum of the moving average. For example with availability_ratio=0.75, a daily maximum
            eight hours average can only be calculated if 18 eight hours average
            are available each of which requires 6 hourly values to be available.
        keep_first_day : bool
            If True, the first day of the period is kept even though
            all of its values are equal to np.nan. If false, the first
            day of the period is dropped.

        Returns
        -------
            Evaluator object with seriesType = 'daily' and with data corresponding to the
            computed daily maximum of the moving average.

        """
        observations = self.reconstruct_observations()
        simulations = self.reconstruct_simulations()
        return self.__class__(observations.movingAverageDailyMax(availability_ratio, keep_first_day),
            simulations.movingAverageDailyMax(availability_ratio, keep_first_day), color=self.color)

    def filteredSeries(self, availability_ratio=0.75):
        """ Return Evaluator object working on the filtered series.

        This method compute the filtered series of observations and simulations
        of the current object and returns a new Evaluator object with
        seriesType = 'hourly' and with data corresponding to the computed
        filtered series.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of values available for a given hour to compute the daily
            filtered series for this hour in all days.

        Returns
        -------
            Evaluator object with seriesType = 'hourly' and with data
            corresponding to the computed filtered series.

        """
        observations = self.reconstruct_observations()
        simulations = self.reconstruct_simulations()
        return self.__class__(observations.filteredSeries(availability_ratio),
            simulations.filteredSeries(availability_ratio), color=self.color)

    def normalized_series(self):
        """ Return Evaluator object with normalized series.

        This method normalizes each series of observations and simulations
        by substracting its median and dividing by its interquartile range.

        Returns
        -------
            Evaluator object with seriesType = 'hourly' and with data
            corresponding to the computed normalized series.

        """
        observations = self.reconstruct_observations()
        simulations = self.reconstruct_simulations()
        return self.__class__(observations.normalized_series(),
            simulations.normalized_series(), color=self.color)

    def contingency_table(self, threshold, outputFile):
        """ Contingency table.

        The table is computed from daily data for each forecast day.
        Before computed the table, values for every stations and every
        days of the period are concatenated. Tables corresponding to
        the different forecast days are stored in the same file.

        Parameters
        ----------
            threshold : scalar
                Threshold value.
            outputFile : str
                File where to save the result.

        """
        if self.seriesType not in ['daily']:
            print("*** Object's attribute 'seriesType' must be 'daily' to use this method !!!")
            return

        with open(outputFile, 'w') as f:
            for fd in range(self.forecastHorizon):
                obs = self.get_obs(fd)
                obs = obs.values.flatten()
                sim = self.get_sim(fd)
                sim = sim.values.flatten()
                res = evt.scores.contingency_table(obs, sim, thr=threshold)
                res = pd.DataFrame(res, index=["obs>s", "obs<s", "total"], columns=["sim>s", "sim<s", "total"]).astype(int)
                # res.to_csv(outputFile.format(forecastDay=fd), sep=' ', na_rep='nan', float_format='%g', header=True, index=True)
                f.write("D{}\n".format(fd) + res.T.to_string() + '\n\n')

    def obs_exceedances(self, threshold, outputFile):
        """ Look for exceedances in observed time series.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        outputFile : str
            File where to save the result.

        """
        ts = self.obsDF.stack()
        idx = ts > threshold
        res = ts.loc[idx]
        with open(outputFile, 'w') as f:
            f.write(res.to_string())

    def sim_exceedances(self, threshold, outputFile):
        """ Look for exceedances in simulated time series.

        Parameters
        ----------
        threshold : scalar
            Threshold value.
        outputFile : str
            File where to save the result. The file name must contain {forecastDay}
            instead of the forecast day number.

        """
        for fd in range(self.forecastHorizon):
            ts = self.get_sim(fd).stack()
            idx = ts > threshold
            res = ts.loc[idx]
            with open(outputFile.format(forecastDay=fd), 'w') as f:
                f.write(res.to_string())
