# coding: utf8
""" This module is designed to compute some Fairmode metrics."""

from __future__ import print_function

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import evaltools as evt
from evaltools.evaluator import Evaluator
from datetime import datetime, date, timedelta, time
from os.path import isdir, isfile


class FairmodeEvaluator(Evaluator):
    """ Class designed to compute some Fairmode metrics.

    This class inherits from evaltools.evaluator.Evaluator.

    Parameters
    ----------
    observations : evaluator.Observations object
        Observations used to construct the Evaluator object.
    simulations : evaluator.Simulations object
        Simulations used to construct the Evaluator object.
    color : str
        Default color that will be used in plotting functions.

    """
    def __init__(self, observations, simulations, color='k'):
        Evaluator.__init__(self, observations, simulations, color)

    # def target_forecast(self, threshold=0.75, forecast_day=0):
    #     """ """
    #     obs = self.get_obs(forecast_day=forecast_day)
    #     intraDayVar = evt.scores.stats2D(obs[1:], obs[:-1], score_list=['RMSE'], axis=0, threshold=threshold, keep_nan=True)

    #     obs = obs[1:]                                      # dropping the 1st time step before calculating RMSE since we need ...
    #     sim = self.get_sim(forecast_day=forecast_day)[1:]  # ... it to calculate the intra-day variation for the 2nd time step
    #     rmse = evt.scores.stats2D(obs, sim, score_list=['RMSE'], axis=0, threshold=threshold, keep_nan=True)

    #     res = rmse/intraDayVar
    #     res.columns = ['target']
    #     return res

    def RMSU(self, U, alpha, RV, threshold=0.75, forecast_day=0):
        """ Root mean square of measurement uncertainty.

        Parameters
        ----------
        U : float
            $U^{RV}_{95,r}$ parameter for meaurement uncertainty calculation.
        alpha : float
            Parameter for meaurement uncertainty calculation.
        RV : float
            Parameter for meaurement uncertainty calculation.
        threshold : int or float
            Minimal number (if type(threshold) is int) or minimal rate (if type(threshold) is float)
            of data available in both obs and sim required to compute the scores.
        forecast_day : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
            pandas.Series with index corresponding to the object stations
            and containing root mean square of measurement uncertainty
            for each station.

        """
        obs = self.get_obs(forecast_day=forecast_day)
        mu = U*np.sqrt((1-alpha**2)*obs**2+(alpha*RV)**2)
        rmsu = evt.scores.stats2D(
            mu, pd.DataFrame(0., index=obs.index, columns=obs.columns),
            score_list=['RMSE'], axis=0, threshold=threshold, keep_nan=True)
        rmsu.columns = ['RMSU']
        return rmsu.RMSU

    def MQI(self, U, alpha, RV, beta=2, threshold=0.75, forecast_day=0):
        """ Modelling quality incator

        Parameters
        ----------
        U : float
            $U^{RV}_{95,r}$ parameter for meaurement uncertainty calculation.
        alpha : float
            Parameter for meaurement uncertainty calculation.
        RV : float
            Parameter for meaurement uncertainty calculation.
        beta : float
            Patameter for MQI calculation.
        threshold : int or float
            Minimal number (if type(threshold) is int) or minimal rate (if type(threshold) is float)
            of data available in both obs and sim required to compute the scores.
        forecast_day : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
            pandas.Series with index corresponding to the object stations
            and containing modelling quality incator for each station.

        """
        obs = self.get_obs(forecast_day=forecast_day)
        sim = self.get_sim(forecast_day=forecast_day)
        rmse = evt.scores.stats2D(obs, sim, score_list=['RMSE'], axis=0, threshold=threshold, keep_nan=True)
        rmsu = self.RMSU(U, alpha, RV, threshold=threshold, forecast_day=forecast_day)
        MQI = rmse.RMSE/(beta*rmsu)
        return MQI

    # def drop_unrepresentative_stations(self, availability_ratio=0.75):
    #     """ Drop stations with a certain rate of missing values.

    #     Modify EvtDataFrame object in place.

    #     Parameters
    #     ----------
    #     availability_ratio : float
    #         Minimal rate of data available on the period required to
    #         keep a station.

    #     """
    #     # keeping
    #     idx = self.obsDF.isna()
    #     for df in self.simDF:
    #         idx = np.logical_or(idx, df.isna())

    #     idx = np.sum(~self.obsDF.isna())/self.obsDF.shape[0] < availability_ratio  # index of stations to be dropped
    #     for df in self.simDF:
    #         idx = np.logical_or(idx, np.sum(~df.isna())/df.shape[0] < availability_ratio)

    #     dropped_sta = idx.index[idx]
    #     self.obsDF.drop(columns=dropped_sta, inplace=True)
    #     for df in self.simDF:
    #         df.drop(columns=dropped_sta, inplace=True)

    def MQI90(self, U, alpha, RV, beta=2, threshold=0.75, forecast_day=0):
        """ 90th percentile of modelling quality indicator

        Parameters
        ----------
        U : float
            $U^{RV}_{95,r}$ parameter for meaurement uncertainty calculation.
        alpha : float
            Parameter for meaurement uncertainty calculation.
        RV : float
            Parameter for meaurement uncertainty calculation.
        beta : float
            Patameter for MQI calculation.
        threshold : int or float
            Minimal number (if type(threshold) is int) or minimal rate (if type(threshold) is float)
            of data available in both obs and sim required to compute the scores.
        forecast_day : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
            Float equal to the 90th percentile of modelling quality incator.

        """
        MQI = self.MQI(U, alpha, RV, beta=beta, threshold=threshold, forecast_day=forecast_day)
        MQI = MQI.dropna().sort_values()
        if len(MQI) == 0:
            return np.nan
        stat90 = int(len(MQI)*0.9)
        dist = len(MQI)*0.9 - stat90
        return MQI.iloc[stat90-1] + (MQI.iloc[stat90] - MQI.iloc[stat90-1])*dist

    def plot_summary(self, availability_ratio=0.75, forecastDay=0, outputFile=None, file_formats=['png'], title=""):
        """ """
        # parameters specific to the species
        if self.species == 'pm10':
            dailyObject = self.dailyMean(availability_ratio=availability_ratio)
            threshold = 50.
            U = 0.28
            alpha = 0.13
            RV = 50
            perc = 0.901
        elif self.species == 'pm2p5':
            dailyObject = self.dailyMean(availability_ratio=availability_ratio)
            threshold = 25
            U = 0.36
            alpha = 0.3
            RV = 25
            perc = 0.901
        elif self.species == 'no2':
            dailyObject = self
            threshold = 200.
            U = 0.24
            alpha = 0.2
            RV = 200
            perc = 0.998
        elif self.species == 'o3':
            dailyObject = self.movingAverageDailyMax(availability_ratio=availability_ratio)
            threshold = 120.
            U = 0.18
            alpha = 0.79
            RV = 120
            perc = 0.929

        perc = 0.95

        # drawing common to all subplots
        def common_params(ax, xmin, xmax, points, mqi=None, sym=True):
            ax.scatter(points, np.ones(len(points)), zorder=10)
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)

            # show outside points
            if (points > xmax).any():
                ax.scatter([xmax+(xmax-xmin)*0.04], [1], clip_on=False, c=dot_col)
            if (points < xmin).any():
                ax.scatter([xmin-(xmax-xmin)*0.04], [1], clip_on=False, c=dot_col)

            plt.tick_params(axis='y', which='both', left=False)
            ax.set_yticklabels([""])
            if sym is True:
                ax.spines['left'].set_visible(False)
                pol = plt.Polygon(xy=[[xmin, 2], [xmin-(xmax-xmin)*0.08, 2], [xmin-(xmax-xmin)*0.08, 0], [xmin, 0]], closed=False, ls='--', clip_on=False, fc='none', edgecolor='k')
                ax.add_patch(pol)
            ax.spines['right'].set_visible(False)
            pol = plt.Polygon(xy=[[xmax, 2], [xmax+(xmax-xmin)*0.08, 2], [xmax+(xmax-xmin)*0.08, 0], [xmax, 0]], closed=False, ls='--', clip_on=False, fc='none', edgecolor='k')
            ax.add_patch(pol)
            ax.tick_params(axis='x', labelsize='small')

            # give more space to y label
            box = ax.get_position()
            ax.set_position([box.x0+box.width*0.15, box.y0, box.width*0.8, box.height*0.4])

            # MQI fulfillment
            if mqi is not None:
                mqo = np.sum(mqi < 1)/float(len(mqi)) >= 0.9
                col = '#4CFF00'*int(mqo) + 'r'*int(~mqo)
                ax.scatter([xmax+(xmax-xmin)*0.15], [1], clip_on=False, c=col, s=100)

        # scores
        scores = dailyObject.stationScores(score_list=['MeanBias', 'PearsonR', 'obs_std', 'sim_std', 'obs_mean', 'sim_mean'],
                                           outputFile=None, availability_ratio=availability_ratio)['D{}'.format(forecastDay)]
        if scores.isna().all().all():
            print("No valid stations !!!")
            return None
        rmsu = dailyObject.RMSU(U=U, alpha=alpha, RV=RV, threshold=float(availability_ratio), forecast_day=forecastDay)
        beta = 2.

        # plotting 
        plt.close('all')
        fig = plt.figure(figsize=(8, 6))
        plt.suptitle(title)
        dot_col = "#1F77B4"
        ymin = 0
        ymax = 2

        # ------- obs mean -------
        ax = fig.add_subplot(811)
        xmin = 0
        xmax = 100

        # obs stats
        av_obs = scores.obs_mean
        ax.set_ylabel("Observed   \nmean ", labelpad=70, rotation='horizontal', verticalalignment='center', size='small')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=av_obs, sym=False)

        ax.text(105, -2, "$\mu gm^{-3}$")
        ax.text(-37, 3, "-- observations --", rotation='vertical')


        # ------- obs exceedences -------
        ax = fig.add_subplot(812)
        xmin = 0
        xmax = 100

        # obs stats
        valid_stations = scores.index[~scores[scores.keys()[0]].isna()]
        nb_exceedences = (dailyObject.get_obs(forecast_day=forecastDay) > threshold)[valid_stations].sum()
        ax.set_ylabel("Observed   \nexceedences \n(> "+str(threshold)+" $\mu gm^{-3}$)", labelpad=70, rotation='horizontal', verticalalignment='center', size='small')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=nb_exceedences, sym=False)

        ax.text(105, -2, "days")

        # ------- Bias -------
        ax = fig.add_subplot(813)
        xmin = -2
        xmax = 2

        rect = plt.Rectangle(xy=(-1.,0.), width=2., height=2., edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(-.7,0.), width=1.4, height=2., edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)
        x = scores['MeanBias']/(beta*rmsu)
        ax.set_ylabel("$\\frac{Mean\\:bias}{\\beta RMS_U}$", labelpad=70, rotation='horizontal', verticalalignment='center', size='large')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=x, mqi=np.abs(x).dropna())

        # ------- Correlation -------
        ax = fig.add_subplot(814)
        xmin = 0
        xmax = 2

        rect = plt.Rectangle(xy=(0.,0.), width=1., height=2., edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(0.,0.), width=.5, height=2., edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)
        # x = scores['PearsonR']/(((beta*rmsu)/(2*scores['obs_std']))**2 - 1)
        # x = (1. - scores['PearsonR']) / (.5*((beta*rmsu)/scores['obs_std'])**2)
        # obs = dailyObject.get_obs(forecast_day=forecast_day)
        # sim = dailyObject.get_sim(forecast_day=forecast_day)
        # stdo = np.nanstd(obs, axis=0)
        # stds = np.nanstd(sim, axis=0)
        # x = np.sqrt(2.*scores['obs_std']*scores['sim_std']*(1. - scores['PearsonR']))/(beta*rmsu)
        x = (2.*scores['obs_std']*scores['sim_std']*(1. - scores['PearsonR']))/(beta*rmsu)**2
        # x = correl
        ax.set_ylabel("$\\frac{2\\sigma_O\\sigma_M(1-R)}{\\beta^2 RMS_U^2}$", labelpad=70, rotation='horizontal', verticalalignment='center', size='large')

        ax.text(-37/50., 7.5, "----------------- time -----------------", rotation='vertical')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=x, mqi=x.dropna(), sym=False)

        # ------- std -------
        ax = fig.add_subplot(815)
        xmin = -2
        xmax = 2

        rect = plt.Rectangle(xy=(-1.,0.), width=2., height=2., edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(-.7,0.), width=1.4, height=2., edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)
        x = (scores['sim_std']-scores['obs_std'])/(beta*rmsu)
        ax.set_yticks([1.])
        ax.set_ylabel("$\\frac{\\sigma_M-\\sigma_O}{\\beta RMS_U}$", labelpad=70, rotation='horizontal', verticalalignment='center', size='large')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=x, mqi=np.abs(x).dropna())

        # ------- Hperc -------
        ax = fig.add_subplot(816)
        xmin = -2
        xmax = 2

        rect = plt.Rectangle(xy=(-1.,0.), width=2., height=2., edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)
        obs = dailyObject.get_obs(forecast_day=forecastDay)
        sim = dailyObject.get_sim(forecast_day=forecastDay)
        perc_values = evt.scores.stats2D(obs, sim, score_list=['obs_percentile', 'sim_percentile'], axis=0, threshold=float(availability_ratio), keep_nan=True, perc=perc)
        U95 = U*np.sqrt((1-alpha**2)*(perc_values['obs_percentile']**2) + (alpha**2)*(RV**2))
        # U95 = U*np.sqrt((1-alpha**2)*(obs.quantile(perc)**2) + (alpha**2)*(RV**2))
        # U95 = U*np.sqrt((1-alpha)*(np.nanstd(perc_values['obs_percentile'])**2 +
        #     np.nanmean(perc_values['obs_percentile'])**2) + alpha*(RV**2))
        # U95 = U*np.sqrt((1-alpha)*(perc_values['obs_percentile']**2) + (alpha)*(RV**2))
        # U95 = U*np.sqrt((1-alpha)*(np.nanstd(obs, axis=0)**2+np.nanmean(obs, axis=0)**2) + (alpha)*(RV**2))
        # U95 = U*np.sqrt((1-alpha**2)*(obs**2) + (alpha**2)*(RV**2))
        # mu =  U*np.sqrt((1-alpha**2)*obs**2   + (alpha*RV)**2)
        # import ipdb
        # ipdb.set_trace()
        # Hperc = (perc_values['sim_percentile'] - perc_values['obs_percentile'])/(beta*U95)
        # Hperc = (sim.quantile(perc) - obs.quantile(perc))/(beta*U95)
        # Hperc = (perc_values['sim_percentile'] - perc_values['obs_percentile'])/(U95)
        Hperc = (perc_values['sim_percentile'] - perc_values['obs_percentile'])/(beta*rmsu)
        # Hperc = (perc_values['sim_percentile'] - perc_values['obs_percentile'])/(beta*np.sqrt(np.nanmean(U95**2)))
        ax.set_ylabel("Hperc", labelpad=70, rotation='horizontal', verticalalignment='center', size='small')
        # ax.set_ylabel("$\\frac{M_{perc}-O_{perc}}{\\beta U_{95}(O_{perc})}$", labelpad=70, rotation='horizontal', verticalalignment='center', size='large')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=Hperc, mqi=np.abs(Hperc).dropna())

        # ------- spatial correlation -------
        ax = fig.add_subplot(817)
        xmin = 0
        xmax = 2

        rect = plt.Rectangle(xy=(0.,0.), width=1., height=2., edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(0.,0.), width=.5, height=2., edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)
        sim =  scores['sim_mean']
        obs = scores['obs_mean']
        # corr = (np.nanmean((obs-np.nanmean(obs))*(sim-np.nanmean(sim))))/(np.nanstd(obs)*np.nanstd(sim))
        N = len(obs)
        corr = ((np.nanmean((obs-np.nanmean(obs))*(sim-np.nanmean(sim))))/(np.nanstd(obs)*np.nanstd(sim)))*(N/(N-1))
        U95 = U*np.sqrt((1-alpha**2)*(obs**2) + (alpha**2)*(RV**2))
        # RMSu_ = np.sqrt(np.nanmean(U95**2))
        RMSu_ = np.sqrt(np.nanmean(rmsu**2))
        # x = corr**2/(beta*RMSu_)**2
        # x = (2.*np.nanstd(obs)*np.nanstd(sim)*(1. - corr))/(beta*RMSu_)**2
        x = (np.nanstd(obs)*np.nanstd(sim)*(1. - corr))/(2.*RMSu_**2)
        ax.set_ylabel("spatial correlation", labelpad=70, rotation='horizontal', verticalalignment='center', size='small')
        # ax.set_ylabel("$\\frac{2\\sigma_\\bar{O}\\sigma_\\bar{M}(1-R)}{\\beta^2 RMS_\\bar{U}^2}$", labelpad=70, rotation='horizontal', verticalalignment='center', size='large')

        ax.text(-37/50., 2, "------ space ------", rotation='vertical')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=np.array([x]), sym=False, mqi=np.array([x]))

        # ------- spatial std -------
        ax = fig.add_subplot(818)
        xmin = -2
        xmax = 2

        rect = plt.Rectangle(xy=(-1.,0.), width=2., height=2., edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(-.7,0.), width=1.4, height=2., edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)
        # x = (np.nanstd(obs)-np.nanstd(sim))/(beta*RMSu_)
        x = (np.nanstd(sim)-np.nanstd(obs))/(beta*RMSu_)
        ax.set_ylabel("spatial std", labelpad=70, rotation='horizontal', verticalalignment='center', size='small')
        ax.annotate('{valid}/{all} valid stations'.format(valid=len(valid_stations), all=len(self.stations)), xy=(1, 0), xycoords='axes fraction', fontsize='large',
                xytext=(40, -20), textcoords='offset points',
                ha='right', va='top')

        # parameters common to all subplots
        common_params(ax, xmin, xmax, points=np.array([x]), mqi=np.abs(np.array([x])))

        # save figure
        if outputFile is None:
            plt.show()
        else:
            for ext in file_formats:
                plt.savefig((outputFile+".{ext}").format(ext=ext))


def plot_targetDiagram(objects, U, alpha, RV, beta=2, availability_ratio=0.75, forecast_day=0,
                       colors=None, title="", outputFile=None, file_formats=['png'],
                       outputCSV=None, list_stations_ouside_target=False):
    """ Target diagram.

    Assessement target diagram as described in FAIRMODE guidance document
    on modelling quality objectives and benchmarking.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    U : float
        $U^{RV}_{95,r}$ parameter for meaurement uncertainty calculation.
    alpha : float
        Parameter for meaurement uncertainty calculation.
    RV : float
        Parameter for meaurement uncertainty calculation.
    beta : float
        Patameter for MQI calculation.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        day to compute the scores for each station.
    forecast_day : int
        Forecast day used in the diagram.
    colors : None or list of str
        Line colors corresponding to each objects.
    title : str
        Diagram title.
    outputFile : str
        File where to save the plot (without extension).
    file_formats : list of str
        List of file extensions.
    outputCSV : str or None
        File where to save the data.

    """
    if colors is None:
        colors = [obj.color for obj in objects]

    plt.close('all')
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)

    # axes
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    major_ticks = np.arange(-2, 3, 1)
    minor_ticks = np.arange(-2, 2.1, 0.1)
    ax.set_xticks(major_ticks)
    ax.set_xticks(minor_ticks, minor=True)
    ax.set_xticklabels([2, 1, 0, 1, 2])
    ax.set_yticks(major_ticks)
    ax.set_yticks(minor_ticks, minor=True)
    ax.yaxis.set_ticks_position('both')
    ax.xaxis.set_ticks_position('both')
    ax.set_aspect('equal')
    plt.xlabel("CRMSE / $\\beta RMS_U$")
    plt.ylabel("Mean Bias / $\\beta RMS_U$")

    # target lines
    for coords in [(-2, 2, -2, 2), (-2, 2, 2, -2)]:  #, (-2, 2, 0, 0), (0, 0, -2, 2)]:
        plt.plot([coords[0], coords[1]], [coords[2], coords[3]], 'k', lw=0.5)
    plt.text(-1.9 , 0, "R", color='k', verticalalignment='center', horizontalalignment='left', size='medium')
    plt.text(1.9 , 0, "SD", color='k', verticalalignment='center', horizontalalignment='right', size='medium')
    plt.text(0, 1.9, "Mean bias > 0", color='k', verticalalignment='top', horizontalalignment='center', size='medium')
    plt.text(0, -1.9, "Mean bias < 0", color='k', verticalalignment='bottom', horizontalalignment='center', size='medium')

    # target circles
    smax = 1
    levels = 2
    xs, ys = np.meshgrid(np.linspace(-smax, smax), np.linspace(-smax, smax))
    rms = np.sqrt(xs**2 + ys**2)
    contours = ax.contour(xs, ys, rms, levels, colors='k', linestyles='dotted', linewidths=0.5)
    circle = plt.Circle((0, 0), 1, color='#ECECEC', zorder=0)
    ax.add_artist(circle)

    # Make place on the side of the figure
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # scatter plot
    for obj, c in zip(objects, colors):
        rmsu = obj.RMSU(U, alpha, RV, threshold=float(availability_ratio), forecast_day=forecast_day)
        obs = obj.get_obs(forecast_day=forecast_day)
        sim = obj.get_sim(forecast_day=forecast_day)
        sc = evt.scores.stats2D(obs, sim, score_list=['MeanBias', 'CRMSE', 'PearsonR'], axis=0, threshold=float(availability_ratio), keep_nan=True)
        if sc.isna().all().all():
            print("No valid station !!!")
            return None
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore', 'Degrees of freedom <= 0 for slice')  # avoid warning when calculating std of all nan vector
            idx_both_nan = np.logical_or(sim.isna(), obs.isna())
            obs=obs.where(~idx_both_nan)
            sim=sim.where(~idx_both_nan)
            stdo = np.std(obs, axis=0)
            stds = np.std(sim, axis=0)
        # signx = (np.abs(stdo - stds)/(stdo*np.sqrt(2*(1-sc.PearsonR))) > 1)*2 - 1
        signx = (np.abs(stdo - stds)/((stdo*stds*2*(1-sc.PearsonR))**0.5) > 1)*2 - 1
        # print((np.abs(stdo - stds)/(stdo*np.sqrt(2*(1-sc['corr'])))).sort_values())
        # signx.sort_values().to_csv('/home/petiotv/fairemode/signx.txt', sep=' ', na_rep='nan', float_format='%g', header=True, index=True)
        x = (sc.CRMSE/(beta*rmsu))*signx
        y = sc.MeanBias/(beta*rmsu)
        ax.scatter(x, y, marker='^', facecolors='none', edgecolors=c,
                   label=obj.model+" (MQI = {})".format(round(obj.MQI90(U, alpha, RV, beta=beta, threshold=float(availability_ratio), forecast_day=forecast_day), 3)))
        if outputCSV is not None:
            radius = np.sqrt(x**2+y**2)
            csv_data = pd.DataFrame({'x': x, 'y': y, 'radius': radius})
            csv_data.sort_values('radius', inplace=True, ascending=False)
            csv_data.to_csv(outputCSV, sep=' ', na_rep='nan', float_format='%g', header=True, index=True)

    # legend
    plt.text(2.1, 1.6, "$\\alpha$ = {}".format(alpha), color='k', verticalalignment='center', horizontalalignment='left', size='medium')
    plt.text(2.1, 1.4, "$\\beta$ = {}".format(beta), color='k', verticalalignment='center', horizontalalignment='left', size='medium')
    plt.text(2.1, 1.2, "RV = {}".format(RV), color='k', verticalalignment='center', horizontalalignment='left', size='medium')
    plt.text(2.1, 1, "$U^{RV}_{95,r}$ = " + "{}".format(U), color='k', verticalalignment='center', horizontalalignment='left', size='medium')
    plt.text(2.1, 0.5, "{} stations".format(np.sum(~x.isna())), color='k', verticalalignment='center', horizontalalignment='left', size='medium')
    # plt.text(2.1, 0.7, "Model : {0}".format(obj.model))
    # plt.text(2.1, 0.5, "Species : {0}".format(obj.species))
    # plt.text(2.1, 0.5, "Period : {0} - {1}".format(obj.startDate.strftime('%Y%m%d'), obj.endDate.strftime('%Y%m%d')))
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    if list_stations_ouside_target is True:
        mqi = np.sqrt(x**2+y**2).sort_values(ascending=False)
        idx = mqi > 1
        out_sta = mqi.loc[idx].index
        sep = [" ", "\n"]
        
        # import ipdb
        # ipdb.set_trace()

        plt.text(2.1, -.25,
                 "{nb} ".format(nb=len(out_sta))+"station"+"s"*(len(out_sta) not in [0, 1])+" with MQI > 1"+(":\n"+''.join([a+sep[i%2] for i,a in enumerate(out_sta[:24])]))*(len(out_sta)>0),
                 color='k', verticalalignment='top', horizontalalignment='left', size='small')

    # titles
    plt.title(title, loc='center')

    # # plt.tight_layout()

    if outputFile is None:
        plt.show()
    else:
        for ext in file_formats:
            plt.savefig((outputFile+".{ext}").format(ext=ext))
