# coding: utf8
""" This module defines Grid, a class designed to interpolate
    values from a grid of values. """

from __future__ import print_function

from datetime import datetime, date, timedelta, time
import numpy as np
import pandas as pd
import evaltools as evt
import matplotlib.pyplot as plt
import os
from six import iteritems


class Grid(object):
    """ Class designed to interpolate data from a lat/lon grid.

    The grid must have equidistant latitudes/longitudes.

    Parameters
    ----------
    minLat : scalar
        Minimal latitude values of the grid.
    minLon : scalar
        Minimal longitude values of the grid.
    dLon : scalar
        Longitude step width of the grid.
    dLat : scalar
        Latitude step width of the grid.
    nbLon : int
        Number of longitude steps in the grid (shape[1] of the grid).
    nbLat : int
        Number of latitude steps in the grid (shape[0] of the grid).

    """
    def __init__(self, minLat, minLon, dLon, dLat, nbLon, nbLat):
        self.minLat = minLat
        self.minLon = minLon
        self.dLon = dLon
        self.dLat = dLat
        self.nbLon = nbLon
        self.nbLat = nbLat
        self.maxLat = minLat + dLat*(nbLat - 1)
        self.maxLon = minLon + dLon*(nbLon - 1)
        self.grids = None
        self.stations = None
        self.nearestPoints = None
        self.stations_to_interpolate = {}

    def load_stations_metadata(self, listing_path, **kwargs):
        """ Load stations coordinate and find nearest grid points.

        Retrieve a list of stations with their coordinates and find
        the four nearest grid points from each of them for a later
        use in the interpolation method.

        The listing file is read with evaltools.utils.read_listing
        and therefore it must be in the correct format.

        Parameters
        ----------
        listing_path : str
            Path of the listing file.
        **kwargs :
            These parameters (like 'sep', 'sub_list', ...) will be
            passed to evaltools.utils.read_listing().

        """
        # read listing
        self.stations = evt.utils.read_listing(listing_path, area_coord=[self.minLon, self.maxLon, self.minLat, self.maxLat], **kwargs)

        # find four nearest grid points
        self.nearestPoints = {}
        for sta in self.stations.index:
            self.nearestPoints[sta] = find_nearest(
                obsLon=self.stations.loc[sta].lon,
                obsLat=self.stations.loc[sta].lat,
                maxLat=self.maxLat,
                minLon=self.minLon,
                dLon=self.dLon,
                dLat=self.dLat,
                nbLon=self.nbLon,
                invertLat=True)

    def load_station_lists(self, station_lists):
        """ Import station lists used for interpolation.

        Parameters
        ----------
        station_lists : dict of 1D arrays
            Dictionary which keys are names for the different grids (species,
            hours, ...) and values are 1D arrays of station codes.

        """
        # check if self.stations has been initialised
        if self.stations is None:
            print("*** self.stations is None !!! it must be initialised with load_stations_metadata object method.")
            return None

        for key, val in iteritems(station_lists):
            idx_in_stations = np.isin(np.array(val), self.stations.index)
            if np.sum(~idx_in_stations) != 0:
                print("*** {} stations dropped since they do not appear in self.stations.index !!!".format(sum(~idx_in_stations)))
            self.stations_to_interpolate[key] = val[idx_in_stations]

    def set_grids(self, grids):
        """ Set grids values from a dictionary of 2D arrays.

        Parameters
        ----------
        grids : dict of 2D arrays
            Dictionary which keys are are names for the different grids
            (species, hours, ...), and values are 2D arrays of float with
            shape corresponding to (self.nbLat, self.nbLon).

        """
        for key, grid in iteritems(grids):
            assert grid.shape == (self.nbLat, self.nbLon), "grids[{key}] do not fit {nbLat}x{nbLon} shape".format(key=key, nbLat=self.nbLat, nbLon=self.nbLon)
        self.grids = grids

    def view(self):
        """ Visualize grid values. """
        if self.grids is not None:
            for key, im in iteritems(self.grids):
                plt.close('all')
                plt.title(key)
                plt.imshow(im)
                plt.axis('off')
                plt.show()
        else:
            print("*** self.grids is None !!!")

    def interpolate(self, use_same_listing=False):
        """ Interpolate values for all loaded grids.


        Parameters
        ----------
        use_same_listing : bool
            If False, for each self.grids keys, a list of
            stations for interpolation must have been submitted with
            Grid.load_station_lists method.
            If True, interpolation is done at every stations
            found in the listing file fed to Grid.load_stations_metadata.
        Returns
        -------
            Dictionary with one key per grid, containing pandas.Series of interpolated values.

        """
        # check if grids and station lists have been initialised
        if self.grids is None:
            print("*** self.grids is None !!!")
            return None
        if use_same_listing is False:
            idx_loaded_lists = np.isin(self.grids.keys(), self.stations_to_interpolate.keys())
            if idx_loaded_lists.all() is False:
                print("*** No station list for grid {} !!! must be initialised " \
                    "with load_station_lists object method.".format(self.grids.keys()[~idx_loaded_lists]))
                return None
        else:
            # check if self.stations has been initialised
            if self.stations is None:
                print("*** self.stations is None !!! it must be initialised with load_stations_metadata object method.")
                return None

        # interpolation
        res = {}
        for key, grid in iteritems(self.grids):
            field = grid.flatten()
            vals = {}
            if use_same_listing is False:
                coords_to_interpolate = self.stations.loc[self.stations_to_interpolate[key]][['lat', 'lon']].itertuples()
            else:
                coords_to_interpolate = self.stations[['lat', 'lon']].itertuples()
            for station, inlat, inlon in coords_to_interpolate:
                points = self.nearestPoints[station]
                p = [(lon, lat, field[idx]) for idx, lat, lon in zip(
                    points["indices"], points["latitudes"], points["longitudes"])]
                vals[station] = bilinear_interpolation(inlon, inlat, p, sort=False)
            res[key] = pd.Series(vals)
        return res


def find_nearest(obsLat, obsLon, maxLat, minLon, dLon, dLat, nbLon, invertLat=True):
    """ Find the 4 nearest grid points for a given lat/lon.

    The grid must have equidistant latitudes/longitudes.

    Returned indices are corresponding to grid data stored in
    a 1D-array (ie the grid is flatten in row-major (C-style) order).

    Parameters
    ----------
    obsLon : float
        Longitude of the input observation.
    obsLat : float
        Latitude of the input observation.
    maxLat : float
        Minimal latitude of the grid.
    minLon : float
        Minimal longitude of the grid.
    dLat : float
        Latitude step width of the grid.
    dLon : float
        Longitude step width of the grid.
    nbLon : int
        Number of longitude steps in the grid (shape[1] of the grid).
    invertLat : bool
        Must be set to True if latitudes are stored in decreasing
        order in the grid.

    Returns
    -------
        Dictionary which keys are 'latitudes', 'longitudes' and 'indices', and values are list of four floats.

    """
    # verify args consistency
    maxLon = minLon + dLon*(nbLon-1)
    if obsLat > maxLat or obsLon < minLon or obsLon > maxLon:
        raise evt.EvaltoolsError('Input point outside the grid !!!')

    # find lat/lon indices of the upper left grid point of the mesh containing the input point
    j = int((1 - invertLat*2)*(obsLat - maxLat)/dLat)
    i = int((obsLon - minLon) / dLon)

    # calculate min/max latitudes of the mesh containing the input point
    modLat_0 = maxLat + (1 - invertLat * 2)*j*dLat
    modLat_1 = modLat_0 + (1 - invertLat * 2)*dLat

    # calculate min/max longitudes of the mesh containing the input point
    modLon_0 = minLon + i*dLon
    modLon_1 = modLon_0 + dLon

    # calculate 1D indices of grid points of the mesh containing the input point
    modIndex_0 = i + j*nbLon
    modIndex_1 = modIndex_0 + 1
    modIndex_2 = i + (j + 1)*nbLon
    modIndex_3 = modIndex_2 + 1

    # result
    modIndex = [modIndex_2, modIndex_0, modIndex_3, modIndex_1]
    modLat = np.round([modLat_1, modLat_0, modLat_1, modLat_0], 5)
    modLon = np.round([modLon_0, modLon_0, modLon_1, modLon_1], 5)
    result = {'latitudes': modLat, 'longitudes': modLon, 'indices': modIndex}

    return result


def bilinear_interpolation(x, y, points, sort=True):
    ''' Interpolate (x,y) from values associated with four points.

    Compute a bilinear interpolation for an (x, y) coordinate
    located inside a rectangle defined by four points.

    Parameters
    ----------
    x, y : float
        Coordinates of the input point.
    points : list of four float triplets
        Four points (x, y, value) forming a rectangle.
        The four points can be in any order.

    '''
    # order points by x, then by y
    if sort is True:
        points = sorted(points)

    (x1, y1, q11), (_x1, y2, q12), (x2, _y1, q21), (_x2, _y2, q22) = points

    # verify args consistency
    if x1 != _x1 or x2 != _x2 or y1 != _y1 or y2 != _y2:
        raise evt.EvaltoolsError('points do not form a rectangle !!!')
    if not x1 <= x <= x2 or not y1 <= y <= y2:
        raise evt.EvaltoolsError('(x, y) not within the rectangle !!!')

    # interpolation
    return (q11 * (x2 - x) * (y2 - y) +
            q21 * (x - x1) * (y2 - y) +
            q12 * (x2 - x) * (y - y1) +
            q22 * (x - x1) * (y - y1)
           ) / ((x2 - x1) * (y2 - y1) + 0.0)
